"""
Landlab component that simulates the transport of bed material
sediment through a 1-D river network, while tracking the resulting changes
in bed material grain size and river bed elevation. Model framework
described in Czuba (2018). Additions include: particle abrasion, variable
active layer thickness (Wong et al., 2007).

.. codeauthor:: Allison Pfeiffer, Katy Barnhart, Jon Czuba, Eric Hutton
"""

import warnings

import numpy as np
import scipy.constants
import xarray as xr

from landlab import Component
from landlab.components import FlowDirectorSteepest
from landlab.data_record import DataRecord
from landlab.grid.network import NetworkModelGrid

_SUPPORTED_TRANSPORT_METHODS = ["WilcockCrowe"]
_SUPPORTED_ACTIVE_LAYER_METHODS = ["WongParker", "GrainSizeDependent", "Constant10cm"]

_REQUIRED_PARCEL_ATTRIBUTES = [
    "time_arrival_in_link",
    "abrasion_rate",
    "density",
    "active_layer",
    "location_in_link",
    "D",
    "volume",
]

_ACTIVE = 1
_INACTIVE = 0

_SAND_SIZE = 0.002
_INIT_ACTIVE_LAYER_THICKNESS = 0.03116362


class NetworkSedimentTransporter(Component):
    """Move sediment parcels on a river network.

    Landlab component that simulates the transport of bed material
    sediment through a 1-D river network, while tracking the resulting changes
    in bed material grain size and river bed elevation. Model framework
    described in Czuba (2018). Additions include: particle abrasion, variable
    active layer thickness (Wong et al., 2007).

    This component cares about units. Its time, length, and mass units are
    seconds, meters, and kilograms, by default. The horizontal unit of the
    grid, and the units of the parameters ``g`` and ``fluid_density`` are
    what specify the component units. In addition, the expressions used
    to calculate the transport have units (Wilcock and Crowe, 2003).

    There is a function that assists in plotting the output of this component.
    It is called :func:`~.plot.plot_network_and_parcels`.  Examples of its usage
    can be found in the NetworkSedimentTransporter notebooks (located in the
    "notebooks" folder).

    Examples
    ----------
    >>> import numpy as np
    >>> from landlab.components import FlowDirectorSteepest, NetworkSedimentTransporter
    >>> from landlab import NetworkModelGrid
    >>> from landlab.data_record import DataRecord

    The :class:`~.NetworkSedimentTransporter` moves "parcels" of sediment down a network
    based on a given flow and a given sediment transport formulation. The river
    network is represented by a landlab :class:`~.NetworkModelGrid`.
    Flow direction in the network is determined using a landlab flow director.
    Sediment parcels are represented as items within a landlab
    :class:`~.DataRecord`. The landlab :class:`~.DataRecord` is used to track
    the location, grain size, sediment density, and total volume of each parcel.

    Create a :class:`~.NetworkModelGrid` to represent the river channel network.
    In this case, the grid is a single line of 4 nodes connected by 3 links. Each
    link represents a reach of river.

    >>> y_of_node = (0, 0, 0, 0)
    >>> x_of_node = (0, 100, 200, 300)
    >>> nodes_at_link = ((0,1), (1,2), (2,3))
    >>> nmg = NetworkModelGrid((y_of_node, x_of_node), nodes_at_link)

    Add required channel and topographic variables to the
    :class:`~.network.NetworkModelGrid`.

    >>> _ = nmg.add_field("bedrock__elevation", [3., 2., 1., 0.], at="node") # m
    >>> _ = nmg.add_field("reach_length", [100., 100., 100.], at="link")  # m
    >>> _ = nmg.add_field(
    ...     "channel_width", (15 * np.ones(nmg.size("link"))), at="link"
    ... )
    >>> _ = nmg.add_field(
    ...     "flow_depth", (2 * np.ones(nmg.size("link"))), at="link"
    ... ) # m

    Add ``"topographic__elevation"`` to the grid because the
    :class:`~.FlowDirectorSteepest` will look to it to
    determine the direction of sediment transport through the network. Each
    time we run the `NetworkSedimentTransporter` the topography will be
    updated based on the bedrock elevation and the distribution of alluvium.

    >>> _ = nmg.add_field(
    ...     "topographic__elevation",
    ...     np.copy(nmg.at_node["bedrock__elevation"]),
    ...     at="node",
    ... )

    Run :class:`~.FlowDirectorSteepest` to determine the direction of sediment
    transport through the network.

    >>> flow_director = FlowDirectorSteepest(nmg)
    >>> flow_director.run_one_step()

    Define the starting time and the number of timesteps for this model run.

    >>> timesteps = 10
    >>> time = [0.0]

    Define the sediment characteristics that will be used to create the parcels
    :class:`~.DataRecord`.

    >>> items = {"grid_element": "link", "element_id": np.array([[0]])}

    >>> variables = {
    ...     "starting_link": (["item_id"], np.array([0])),
    ...     "abrasion_rate": (["item_id"], np.array([0])),
    ...     "density": (["item_id"], np.array([2650])),
    ...     "time_arrival_in_link": (["item_id", "time"], np.array([[0]])),
    ...     "active_layer": (["item_id", "time"], np.array([[1]])),
    ...     "location_in_link": (["item_id", "time"], np.array([[0]])),
    ...     "D": (["item_id", "time"], np.array([[0.05]])),
    ...     "volume": (["item_id", "time"], np.array([[1]])),
    ... }

    Create the sediment parcel :class:`~.DataRecord`. In this case, we are creating
    a single sediment parcel with all of the required attributes.

    >>> one_parcel = DataRecord(
    ...     nmg,
    ...     items=items,
    ...     time=time,
    ...     data_vars=variables,
    ...     dummy_elements={"link": [NetworkSedimentTransporter.OUT_OF_NETWORK]},
    ... )

    Instantiate the model run

    >>> nst = NetworkSedimentTransporter(
    ...     nmg,
    ...     one_parcel,
    ...     flow_director,
    ...     bed_porosity=0.03,
    ...     g=9.81,
    ...     fluid_density=1000,
    ...     transport_method="WilcockCrowe",
    ...     active_layer_method="WongParker"
    ... )

    >>> dt = 60  # (seconds) 1 min timestep

    Run the model

    >>> for t in range(0, (timesteps * dt), dt):
    ...     nst.run_one_step(dt)

    We can the link location of the parcel at each timestep

    >>> print(one_parcel.dataset.element_id.values)
    [[ 0.  0.  0.  0.  0.  1.  1.  1.  1.  1.  2.]]

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    Pfeiffer, A., Barnhart, K., Czuba, J., & Hutton, E. (2020).
    NetworkSedimentTransporter: A Landlab component for bed material transport
    through river networks. Journal of Open Source Software, 5(53).

    **Additional References**

    Czuba, J. A. (2018). A Lagrangian framework for exploring complexities
    of mixed-size sediment transport in gravel-bedded river networks.
    Geomorphology, 321, 146-152.

    Wilcock, P. R., & Crowe, J. C. (2003). Surface-based transport model
    for mixed-size sediment. Journal of Hydraulic Engineering, 129(2), 120-128.

    Wong, M., Parker, G., DeVries, P., Brown, T. M., & Burges, S. J. (2007).
    Experiments on dispersion of tracer stones under lower‐regime plane‐bed
    equilibrium bed load transport. Water Resources Research, 43(3).
    """

    _name = "NetworkSedimentTransporter"

    _unit_agnostic = False

    __version__ = "1.0"

    _info = {
        "bedrock__elevation": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "elevation of the bedrock surface",
        },
        "channel_slope": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m/m",
            "mapping": "link",
            "doc": "Slope of the river channel through each reach",
        },
        "channel_width": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "link",
            "doc": "Flow width of the channel, assuming constant width",
        },
        "flow_depth": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "link",
            "doc": "Flow depth of the channel",
        },
        "reach_length": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "link",
            "doc": "Length of each reach",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
    }

    #: Indicates a parcel is out of network
    OUT_OF_NETWORK = NetworkModelGrid.BAD_INDEX - 1

    def __init__(
        self,
        grid,
        parcels,
        flow_director,
        bed_porosity=0.3,
        g=scipy.constants.g,
        fluid_density=1000.0,
        transport_method="WilcockCrowe",
        active_layer_method="WongParker",
        active_layer_d_multiplier=2,
        slope_threshold=1e-4,
    ):
        """
        Parameters
        ----------
        grid: NetworkModelGrid
            A :class:`~.NetworkModelGrid` in which links are stream channel segments.
        parcels: DataRecord
            A landlab :class:`~.DataRecord` describing the characteristics and
            location of sediment "parcels".  At any given timestep, each parcel is
            located at a specified point along (location_in_link) a particular link
            (element_id). Each parcel has a total sediment volume (volume), sediment
            grain size (D), sediment density (density), and bed material abrasion rate
            (abrasion_rate). During a given timestep, parcels may be in the
            "active layer" of most recently deposited sediment
            (active_layer = 1), or they may be buried and not subject to
            transport (active_layer = 0). Whether a sediment parcel is active
            or not is determined based on flow conditions and parcel attributes
            in :meth:`~.NetworkSedimentTransporter.run_one_step`.
        flow_director: :class:`~.FlowDirectorSteepest`
            A landlab flow director. Currently, must be :class:`~.FlowDirectorSteepest`.
        bed_porosity: float, optional
            Proportion of void space between grains in the river channel bed.
        g: float, optional
            Acceleration due to gravity [m / s^2].
        fluid_density: float, optional
            Density of the fluid (generally, water) in which sediment is
            moving [kg / m^3].
        transport_method: {"WilcockCrowe"}, optional
            Sediment transport equation option.
        active_layer_method: {"WongParker", "GrainSizeDependent", "Constant10cm"}, optional
            Option for treating sediment active layer as a constant or variable.
        slope_threshold: float, optional
            Minimum channel slope at any given link. Slopes lower than this
            value will default to the threshold.
        """
        if not isinstance(grid, NetworkModelGrid):
            raise ValueError("grid must be NetworkModelGrid")

        # run super. this will check for required inputs specified by _info
        super().__init__(grid)

        # check key information about the parcels, including that all required
        # attributes are present.
        if not isinstance(parcels, DataRecord):
            raise ValueError("parcels must be an instance of DataRecord")

        for rpa in _REQUIRED_PARCEL_ATTRIBUTES:
            if rpa not in parcels.dataset:
                raise ValueError(f"{rpa} must be assigned to the parcels")

        # save key information about the parcels.
        self._parcels = parcels
        self._num_parcels = self._parcels.number_of_items
        self._time_variable_parcel_attributes = [
            "time_arrival_in_link",
            "active_layer",
            "location_in_link",
            "D",
            "volume",
        ]

        # assert that the flow director is a component and is of type
        # FlowDirectorSteepest
        if not isinstance(flow_director, FlowDirectorSteepest):
            raise ValueError("flow_director must be FlowDirectorSteepest.")

        # save reference to flow director
        self._fd = flow_director

        # verify and save the bed porosity.
        if not 0 <= bed_porosity < 1:
            raise ValueError(f"bed_porosity must be between 0 and 1 ({bed_porosity})")
        self._bed_porosity = bed_porosity

        # save or create other key properties.
        self._g = g
        self._fluid_density = fluid_density
        self._time_idx = 0
        self._time = 0.0
        self._distance_traveled_cumulative = np.zeros(self._num_parcels)
        self._slope_threshold = slope_threshold

        # check the transport method is valid.
        if transport_method in _SUPPORTED_TRANSPORT_METHODS:
            self._transport_method = transport_method
        else:
            raise ValueError(
                f"{transport_method}: Valid transport method not supported."
            )

        # update the update_transport_time function to be the correct function
        # for the transport method.
        if self._transport_method == "WilcockCrowe":
            self._update_transport_time = self._calc_transport_wilcock_crowe

        if active_layer_method in _SUPPORTED_ACTIVE_LAYER_METHODS:
            self._active_layer_method = active_layer_method
        else:
            raise ValueError(
                f"{active_layer_method}: Active layer method not supported."
            )

        if self._active_layer_method == "GrainSizeDependent":
            self._active_layer_d_multiplier = active_layer_d_multiplier

        # save reference to key fields
        self._width = self._grid.at_link["channel_width"]
        self._topographic__elevation = self._grid.at_node["topographic__elevation"]

        # create field for channel_slope and topographic__elevation if they
        # don't yet exist.
        self.initialize_output_fields()
        self._channel_slope = self._grid.at_link["channel_slope"]

        # Adjust topographic elevation based on the parcels present.
        # Note that at present FlowDirector is just used for network connectivity.
        # get alluvium depth and calculate topography from br+alluvium, then update slopes.

        self._create_new_parcel_time()
        self._calculate_mean_D_and_rho()

        self._partition_active_and_storage_layers()
        self._adjust_node_elevation()
        self._update_channel_slopes()

    @property
    def time(self):
        """Return current time."""
        return self._time

    @property
    def d_mean_active(self):
        """Mean parcel grain size of active parcels aggregated at link."""
        return self._d_mean_active

    @property
    def rhos_mean_active(self):
        """Mean parcel density of active parcels aggregated at link."""
        return self._rhos_mean_active

    def _create_new_parcel_time(self):
        """If we are going to track parcels through time in
        :class:`~.DataRecord`, we need to add a new time column to the parcels
        dataframe. This method simply copies over the attributes of the parcels
        from the former timestep.  Attributes will be updated over the course of
        this step.
        """

        if self._time_idx != 0:
            self._parcels.add_record(time=[self._time])

            self._parcels.ffill_grid_element_and_id()

            # copy parcel attributes forward in time.
            for at in self._time_variable_parcel_attributes:
                self._parcels.dataset[at].values[
                    :, self._time_idx
                ] = self._parcels.dataset[at].values[:, self._time_idx - 1]

        self._this_timesteps_parcels = np.zeros_like(
            self._parcels.dataset.element_id, dtype=bool
        )
        self._this_timesteps_parcels[:, -1] = True

        parcels_off_grid = (
            self._parcels.dataset.element_id[:, -1] == self.OUT_OF_NETWORK
        )
        self._this_timesteps_parcels[parcels_off_grid, -1] = False

        self._num_parcels = self._parcels.number_of_items
        # ^ needs to run just in case we've added more parcels

    def _update_channel_slopes(self):
        """Re-calculate channel slopes during each timestep."""

        for i in range(self._grid.number_of_links):
            upstream_node_id = self._fd.upstream_node_at_link()[i]
            downstream_node_id = self._fd.downstream_node_at_link()[i]

            self._channel_slope[i] = _recalculate_channel_slope(
                self._grid.at_node["topographic__elevation"][upstream_node_id],
                self._grid.at_node["topographic__elevation"][downstream_node_id],
                self._grid.at_link["reach_length"][i],
                self._slope_threshold,
            )

    def _calculate_mean_D_and_rho(self):
        """Calculate mean grain size and density on each link"""

        current_parcels = self._parcels.dataset.isel(time=self._time_idx)

        # In the first full timestep, we need to calc grain size & rho_sed.
        # Assume all parcels are in the active layer for the purposes of
        # grain size and mean sediment density calculations

        # FUTURE: make it possible to circumvent this if mean grain size
        # has already been calculated (e.g. during 'zeroing' runs)

        # Calculate mean values for density and grain size (weighted by volume).
        sel_parcels = current_parcels.where(
            current_parcels.element_id != self.OUT_OF_NETWORK
        )

        d_weighted = sel_parcels.D * sel_parcels.volume
        rho_weighted = sel_parcels.density * sel_parcels.volume
        d_weighted.name = "d_weighted"
        rho_weighted.name = "rho_weighted"

        grouped_by_element = xr.merge(
            (sel_parcels.element_id, sel_parcels.volume, d_weighted, rho_weighted)
        ).groupby("element_id")

        d_avg = grouped_by_element.sum().d_weighted / grouped_by_element.sum().volume
        rho_avg = (
            grouped_by_element.sum().rho_weighted / grouped_by_element.sum().volume
        )

        self._d_mean_active = np.zeros(self._grid.size("link"))
        self._d_mean_active[d_avg.element_id.values.astype(int)] = d_avg.values

        self._rhos_mean_active = np.zeros(self._grid.size("link"))
        self._rhos_mean_active[rho_avg.element_id.values.astype(int)] = rho_avg.values

    def _partition_active_and_storage_layers(self, **kwds):
        """For each parcel in the network, determines whether it is in the
        active or storage layer during this timestep, then updates node
        elevations.
        """
        self._vol_tot = self._parcels.calc_aggregate_value(
            xr.Dataset.sum,
            "volume",
            at="link",
            filter_array=self._this_timesteps_parcels,
            fill_value=0.0,
        )

        if self._active_layer_method == "WongParker":
            # Wong et al. (2007) approximation for active layer thickness.
            # NOTE: calculated using grain size and grain density calculated for
            # the active layer grains in each link at the **previous** timestep.
            # This circumvents the need for an iterative scheme to determine grain
            # size of the active layer before determining which grains are in the
            # active layer.

            # calculate tau
            tau = (
                self._fluid_density
                * self._g
                * self._grid.at_link["channel_slope"]
                * self._grid.at_link["flow_depth"]
            )

            # calcuate taustar
            # taustar = tau / (
            #     (self._rhos_mean_active - self._fluid_density)
            #     * self._g
            #     * self._d_mean_active
            # )
            taustar = np.zeros_like(tau)
            np.divide(
                tau,
                (self._rhos_mean_active - self._fluid_density)
                * self._g
                * self._d_mean_active,
                where=self._rhos_mean_active > self._fluid_density,
                out=taustar,
            )

            # calculate active layer thickness
            self._active_layer_thickness = (
                0.515
                * self._d_mean_active
                * (3.09 * (taustar - 0.0549).clip(0.0, None) ** 0.56)
            )  # in units of m

        elif self._active_layer_method == "GrainSizeDependent":
            # Set all active layers to a multiple of the lnk mean grain size
            self._active_layer_thickness = (
                self._d_mean_active * self._active_layer_d_multiplier
            )

        elif self._active_layer_method == "Constant10cm":
            # Set all active layers to 10 cm thickness.
            self._active_layer_thickness = 0.1 * np.ones_like(self._d_mean_active)

        # If links have no parcels, we still need to assign them an active layer
        # thickness..
        links_with_no_active_layer = np.isnan(self._active_layer_thickness)
        self._active_layer_thickness[links_with_no_active_layer] = np.mean(
            self._active_layer_thickness[links_with_no_active_layer == 0]
        )  # assign links with no parcels an average value

        if np.sum(np.isfinite(self._active_layer_thickness)) == 0:
            self._active_layer_thickness.fill(_INIT_ACTIVE_LAYER_THICKNESS)
            # handles the case of the first timestep -- assigns a modest value

        capacity = (
            self._grid.at_link["channel_width"]
            * self._grid.at_link["reach_length"]
            * self._active_layer_thickness
        )  # in units of m^3

        active_inactive = _INACTIVE * np.ones(self._num_parcels)

        current_link = self._parcels.dataset.element_id.values[:, -1].astype(int)
        time_arrival = self._parcels.dataset.time_arrival_in_link.values[:, -1]
        volumes = self._parcels.dataset.volume.values[:, -1]

        for i in range(self._grid.number_of_links):
            if (
                self._vol_tot[i] > 0
            ):  # only do this check capacity if parcels are in link
                # First In Last Out.

                # Find parcels on this link.
                this_links_parcels = np.where(current_link == i)[0]

                # sort them by arrival time.
                time_arrival_sort = np.flip(
                    np.argsort(
                        time_arrival[this_links_parcels],
                        0,
                    )
                )
                parcel_id_time_sorted = this_links_parcels[time_arrival_sort]

                # calculate the cumulative volume (in sorted order).
                cumvol = np.cumsum(volumes[parcel_id_time_sorted])

                # determine which parcels are within capacity and set those to
                # active.
                make_active = parcel_id_time_sorted[cumvol <= capacity[i]]

                active_inactive[make_active] = _ACTIVE

        self._parcels.dataset.active_layer[:, -1] = active_inactive

        # set active here. reference it below in wilcock crowe
        self._active_parcel_records = (
            self._parcels.dataset.active_layer == _ACTIVE
        ) * (self._this_timesteps_parcels)

        self._vol_act = self._parcels.calc_aggregate_value(
            xr.Dataset.sum,
            "volume",
            at="link",
            filter_array=self._active_parcel_records,
            fill_value=0.0,
        )

        self._vol_stor = (self._vol_tot - self._vol_act) / (1 - self._bed_porosity)

    def _adjust_node_elevation(self):
        """Adjusts slope for each link based on parcel motions from last
        timestep and additions from this timestep.
        """

        number_of_contributors = np.sum(
            self._fd.flow_link_incoming_at_node() == 1, axis=1
        )
        downstream_link_id = self._fd.link_to_flow_receiving_node
        # USED TO BE      downstream_link_id = self._fd.link_to_flow_receiving_node[
        #            self._fd.downstream_node_at_link()
        #        ]
        upstream_contributing_links_at_node = np.where(
            self._fd.flow_link_incoming_at_node() == 1, self._grid.links_at_node, -1
        )

        # Update the node topographic elevations depending on the quantity of stored sediment
        for n in range(self._grid.number_of_nodes):
            if number_of_contributors[n] > 0:  # we don't update head node elevations
                upstream_links = upstream_contributing_links_at_node[n]
                real_upstream_links = upstream_links[
                    upstream_links != self._grid.BAD_INDEX
                ]
                width_of_upstream_links = self._grid.at_link["channel_width"][
                    real_upstream_links
                ]
                length_of_upstream_links = self._grid.at_link["reach_length"][
                    real_upstream_links
                ]

                #                ALERT: Moved this to the "else" statement below. AP 11/11/19
                #                length_of_downstream_link = self._grid.at_link["reach_length"][
                #                    downstream_link_id
                #                ][n]
                #                width_of_downstream_link = self._grid.at_link["channel_width"][
                #                    downstream_link_id
                #                ][n]

                if (
                    downstream_link_id[n] == self._grid.BAD_INDEX
                ):  # I'm sure there's a better way to do this, but...
                    length_of_downstream_link = 0
                    width_of_downstream_link = 0
                else:
                    length_of_downstream_link = self._grid.at_link["reach_length"][
                        downstream_link_id
                    ][n]
                    width_of_downstream_link = self._grid.at_link["channel_width"][
                        downstream_link_id
                    ][n]

                alluvium__depth = _calculate_alluvium_depth(
                    self._vol_stor[downstream_link_id][n],
                    width_of_upstream_links,
                    length_of_upstream_links,
                    width_of_downstream_link,
                    length_of_downstream_link,
                    self._bed_porosity,
                )

                self._grid.at_node["topographic__elevation"][n] = (
                    self._grid.at_node["bedrock__elevation"][n] + alluvium__depth
                )

    def _calc_transport_wilcock_crowe(self):
        """Method to determine the transport time for each parcel in the active
        layer using a sediment transport equation.

        Note: could have options here (e.g. Wilcock and Crowe, FLVB, MPM, etc)
        """
        # Initialize _pvelocity, the virtual velocity of each parcel
        # (link length / link travel time)
        self._pvelocity = np.zeros(self._num_parcels)

        # parcel attribute arrays from DataRecord

        Darray = self._parcels.dataset.D[:, self._time_idx]
        Activearray = self._parcels.dataset.active_layer[:, self._time_idx].values
        Rhoarray = self._parcels.dataset.density.values
        Volarray = self._parcels.dataset.volume[:, self._time_idx].values
        Linkarray = self._parcels.dataset.element_id[
            :, self._time_idx
        ].values  # link that the parcel is currently in

        R = (Rhoarray - self._fluid_density) / self._fluid_density

        # parcel attribute arrays to populate below
        frac_sand_array = np.zeros(self._num_parcels)
        vol_act_array = np.zeros(self._num_parcels)
        Sarray = np.zeros(self._num_parcels)
        Harray = np.zeros(self._num_parcels)
        Larray = np.zeros(self._num_parcels)
        # D_mean_activearray = np.zeros(self._num_parcels) * (np.nan)
        # active_layer_thickness_array = np.zeros(self._num_parcels) * np.nan
        D_mean_activearray = np.full(self._num_parcels, np.nan)
        active_layer_thickness_array = np.full(self._num_parcels, np.nan)

        #        rhos_mean_active = np.zeros(self._num_parcels)
        #        rhos_mean_active.fill(np.nan)

        # find active sand
        # since find active already sets all prior timesteps to False, we
        # can use D for all timesteps here.
        findactivesand = (
            self._parcels.dataset.D < _SAND_SIZE
        ) * self._active_parcel_records

        vol_act_sand = self._parcels.calc_aggregate_value(
            xr.Dataset.sum,
            "volume",
            at="link",
            filter_array=findactivesand,
            fill_value=0.0,
        )

        frac_sand = np.zeros_like(self._vol_act)
        frac_sand[self._vol_act != 0.0] = (
            vol_act_sand[self._vol_act != 0.0] / self._vol_act[self._vol_act != 0.0]
        )
        frac_sand[np.isnan(frac_sand)] = 0.0

        # Calc attributes for each link, map to parcel arrays
        for i in range(self._grid.number_of_links):
            active_here = np.nonzero(
                np.logical_and(Linkarray == i, Activearray == _ACTIVE)
            )[0]
            d_act_i = Darray[active_here]
            vol_act_i = Volarray[active_here]
            rhos_act_i = Rhoarray[active_here]
            vol_act_tot_i = np.sum(vol_act_i)

            self._d_mean_active[i] = np.sum(d_act_i * vol_act_i) / (vol_act_tot_i)
            if vol_act_tot_i > 0:
                self._rhos_mean_active[i] = np.sum(rhos_act_i * vol_act_i) / (
                    vol_act_tot_i
                )
            else:
                self._rhos_mean_active[i] = np.nan
            D_mean_activearray[Linkarray == i] = self._d_mean_active[i]
            frac_sand_array[Linkarray == i] = frac_sand[i]
            vol_act_array[Linkarray == i] = self._vol_act[i]
            Sarray[Linkarray == i] = self._grid.at_link["channel_slope"][i]
            Harray[Linkarray == i] = self._grid.at_link["flow_depth"][i]
            Larray[Linkarray == i] = self._grid.at_link["reach_length"][i]
            active_layer_thickness_array[Linkarray == i] = self._active_layer_thickness[
                i
            ]

        # Wilcock and Crowe calculate transport for all parcels (active and inactive)
        taursg = _calculate_reference_shear_stress(
            self._fluid_density, R, self._g, D_mean_activearray, frac_sand_array
        )

        frac_parcel = np.nan * np.zeros_like(Volarray)
        frac_parcel[vol_act_array != 0.0] = (
            Volarray[vol_act_array != 0.0] / vol_act_array[vol_act_array != 0.0]
        )

        b = 0.67 / (1.0 + np.exp(1.5 - Darray / D_mean_activearray))

        tau = self._fluid_density * self._g * Harray * Sarray
        tau = np.atleast_1d(tau)

        taur = taursg * (Darray / D_mean_activearray) ** b
        tautaur = tau / taur
        tautaur_cplx = tautaur.astype(np.complex128)
        # ^ work around needed b/c np fails with non-integer powers of negative numbers

        W = 0.002 * np.power(tautaur_cplx.real, 7.5)
        W[tautaur >= 1.35] = 14 * np.power(
            (1 - (0.894 / np.sqrt(tautaur_cplx.real[tautaur >= 1.35]))), 4.5
        )

        active_parcel_idx = Activearray == _ACTIVE

        # compute parcel virtual velocity, m/s
        self._pvelocity[active_parcel_idx] = (
            W.real[active_parcel_idx]
            * (tau[active_parcel_idx] ** (3.0 / 2.0))
            * frac_parcel[active_parcel_idx]
            / (self._fluid_density ** (3.0 / 2.0))
            / self._g
            / R[active_parcel_idx]
            / active_layer_thickness_array[active_parcel_idx]
        )

        self._pvelocity[np.isnan(self._pvelocity)] = 0.0

        if np.max(self._pvelocity) > 1:
            warnings.warn(
                "NetworkSedimentTransporter: Maximum parcel virtual velocity exceeds 1 m/s",
                stacklevel=2,
            )

        # Assign those things to the grid -- might be useful for plotting
        self._grid.at_link["sediment_total_volume"] = self._vol_tot
        self._grid.at_link["sediment__active__volume"] = self._vol_act
        self._grid.at_link["sediment__active__sand_fraction"] = frac_sand

    def _move_parcel_downstream(self, dt):
        """Method to update parcel location for each parcel in the active layer."""
        # determine where parcels are starting
        current_link = self._parcels.dataset.element_id.values[:, -1].astype(int)
        self.current_link = current_link

        # determine location within link where parcels are starting.
        location_in_link = self._parcels.dataset.location_in_link.values[:, -1]

        # determine how far each parcel needs to travel this timestep.
        distance_to_travel_this_timestep = self._pvelocity * dt
        # total distance traveled in dt at parcel virtual velocity
        # Note: movement in current and any DS links at this dt is at the same
        # velocity as in the current link perhaps modify in the future

        # Accumulate the total distance traveled by a parcel for abrasion rate
        # calculations.
        if np.size(self._distance_traveled_cumulative) != np.size(
            distance_to_travel_this_timestep
        ):
            dist_array = distance_to_travel_this_timestep
            dist_array[: self._num_parcels] += distance_to_travel_this_timestep
            self._distance_traveled_cumulative = dist_array
        else:
            self._distance_traveled_cumulative += distance_to_travel_this_timestep
            # ^ accumulates total distanced traveled for testing abrasion

        # active parcels on the network:
        in_network = (
            self._parcels.dataset.element_id.values[:, self._time_idx]
            != self.OUT_OF_NETWORK
        )
        active = distance_to_travel_this_timestep > 0.0
        active_parcel_ids = np.nonzero(in_network * active)[0]

        distance_left_to_travel = distance_to_travel_this_timestep.copy()
        while np.any(distance_left_to_travel > 0.0):
            # Step 1: Move parcels downstream.
            on_network = current_link != self.OUT_OF_NETWORK

            # Get current link lengths:
            current_link_lengths = self._grid.at_link["reach_length"][current_link]

            # Determine where they are in the current link.
            distance_to_exit_current_link = current_link_lengths * (
                1.0 - location_in_link
            )

            # Identify which ones will come to rest in the current link.
            rest_this_link = (
                (distance_left_to_travel < distance_to_exit_current_link)
                * on_network
                * (distance_left_to_travel > 0.0)
            )

            # Deal with those staying in the current link.
            if np.any(rest_this_link):
                # print('  {x} coming to rest'.format(x=np.sum(rest_this_link)))

                # for those staying in this link, calculate the location in link
                # (note that this is a proportional distance). AND change
                # distance_left_to_travel to 0.0
                location_in_link[rest_this_link] = 1.0 - (
                    (
                        distance_to_exit_current_link[rest_this_link]
                        - distance_left_to_travel[rest_this_link]
                    )
                    / current_link_lengths[rest_this_link]
                )

                distance_left_to_travel[rest_this_link] = 0.0

            # Deal with those moving to a downstream link.
            moving_downstream = (
                (distance_left_to_travel >= distance_to_exit_current_link)
                * on_network
                * (distance_left_to_travel > 0.0)
            )
            if np.any(moving_downstream):
                # print('  {x} next link'.format(x=np.sum(moving_downstream)))
                # change location in link to 0
                location_in_link[moving_downstream] = 0.0

                # decrease distance to travel.
                distance_left_to_travel[
                    moving_downstream
                ] -= distance_to_exit_current_link[moving_downstream]

                # change current link to the downstream link.

                # get the downstream link at link:
                downstream_node = self._fd.downstream_node_at_link()[current_link]
                downstream_link = self._fd.link_to_flow_receiving_node[downstream_node]

                # assign new values to current link.
                current_link[moving_downstream] = downstream_link[moving_downstream]

                # find and address those links who have moved out of network.
                moved_oon = downstream_link == self._grid.BAD_INDEX

                if np.any(moved_oon):
                    # print('  {x} exiting network'.format(x=np.sum(moved_oon)))

                    current_link[moved_oon] = self.OUT_OF_NETWORK
                    # assign location in link of np.nan to those which moved oon
                    location_in_link[moved_oon] = np.nan
                    distance_left_to_travel[moved_oon] = 0.0

        # Step 2: Parcel is at rest... Now update its information.

        # reduce D and volume due to abrasion
        vol = _calculate_parcel_volume_post_abrasion(
            self._parcels.dataset.volume[active_parcel_ids, self._time_idx],
            distance_to_travel_this_timestep[active_parcel_ids],
            self._parcels.dataset.abrasion_rate[active_parcel_ids],
        )

        D = _calculate_parcel_grain_diameter_post_abrasion(
            self._parcels.dataset.D[active_parcel_ids, self._time_idx],
            self._parcels.dataset.volume[active_parcel_ids, self._time_idx],
            vol,
        )

        # update parcel attributes

        # arrival time in link
        self._parcels.dataset.time_arrival_in_link[
            active_parcel_ids, self._time_idx
        ] = self._time_idx

        # location in link
        self._parcels.dataset.location_in_link[
            active_parcel_ids, self._time_idx
        ] = location_in_link[active_parcel_ids]

        self._parcels.dataset.element_id[
            active_parcel_ids, self._time_idx
        ] = current_link[active_parcel_ids]
        #                self._parcels.dataset.active_layer[p, self._time_idx] = 1
        # ^ reset to 1 (active) to be recomputed/determined at next timestep
        self._parcels.dataset.D[active_parcel_ids, self._time_idx] = D
        self._parcels.dataset.volume[active_parcel_ids, self._time_idx] = vol

    def run_one_step(self, dt):
        """Run :class:`~.NetworkSedimentTransporter` forward in time.

        When the :class:`~.NetworkSedimentTransporter` runs forward in time the following
        steps occur:

        1. A new set of records is created in the Parcels that corresponds to the new time
        2. If parcels are on the network then:

           a. Active parcels are identified based on entrainment critera.
           b. Effective bed slope is calculated based on inactive parcel volumes.
           c. Transport rate is calculated.
           d. Active parcels are moved based on the tranport rate.

        Parameters
        ----------
        dt : float
            Duration of time to run the :class:`~.NetworkSedimentTransporter` forward.

        Raises
        ------
        RuntimeError
            If no parcels remain on the grid.

        """
        self._time += dt
        self._time_idx += 1
        self._create_new_parcel_time()

        if self._this_timesteps_parcels.any():
            self._partition_active_and_storage_layers()
            self._adjust_node_elevation()
            self._update_channel_slopes()
            self._update_transport_time()
            self._move_parcel_downstream(dt)

        else:
            raise RuntimeError("No more parcels on grid")


# %% Methods referenced above, separated for purposes of testing


def _recalculate_channel_slope(z_up, z_down, dx, threshold):
    """Recalculate channel slope based on elevation.

    Parameters
    ----------
    z_up : float
        Upstream elevation.
    z_down : float
        Downstream elevation.
    dz : float
        Distance.
    threshold : float
        Minimum channel slope threshold.

    Examples
    --------
    >>> import pytest

    >>> _recalculate_channel_slope(10., 0., 10., 0.0001)
    1.0
    >>> _recalculate_channel_slope(0., 0., 10.,0.0001)
    0.0001
    >>> with pytest.warns(UserWarning):
    ...     _recalculate_channel_slope(0., 10., 10.,0.0001)
    0.0
    """
    chan_slope = (z_up - z_down) / dx

    if chan_slope < 0.0:
        chan_slope = 0.0
        warnings.warn(
            "NetworkSedimentTransporter: Negative channel slope encountered.",
            UserWarning,
            stacklevel=2,
        )

    elif chan_slope < threshold:
        chan_slope = threshold

    return chan_slope


def _calculate_alluvium_depth(
    stored_volume,
    width_of_upstream_links,
    length_of_upstream_links,
    width_of_downstream_link,
    length_of_downstream_link,
    porosity,
):
    """Calculate alluvium depth based on adjacent link inactive parcel volumes.

    Parameters
    ----------
    stored_volume : float
        Total volume of inactive parcels in this link.
    width_of_upstream_links : float
        Channel widths of upstream links.
    length_of_upstream_link : float
        Channel lengths of upstream links.
    width_of_downstream_link : float
        Channel widths of downstream links.
    length_of_downstream_link : float
        Channel lengths of downstream links.
    porosity: float
        Channel bed sediment porosity.

    Examples
    --------
    >>> import pytest
    >>> _calculate_alluvium_depth(100,np.array([0.5,1]),np.array([10,10]), 1, 10, 0.2)
    10.0
    >>> _calculate_alluvium_depth(24,np.array([0.1,3]),np.array([10,10]), 1, 1, 0.5)
    3.0
    >>> with pytest.raises(ValueError):
    ...     _calculate_alluvium_depth(24,np.array([0.1,3]),np.array([10,10]), 1, 1, 2)
    """

    alluvium__depth = (
        2
        * stored_volume
        / (
            np.sum(width_of_upstream_links * length_of_upstream_links)
            + width_of_downstream_link * length_of_downstream_link
        )
        / (1 - porosity)
    )

    if alluvium__depth < 0.0:
        raise ValueError("NST Alluvium Depth Negative")

    return alluvium__depth


def _calculate_reference_shear_stress(
    fluid_density, R, g, mean_active_grain_size, frac_sand
):
    """Calculate reference Shields stress (taursg) using the sand content of
    the bed surface, as per Wilcock and Crowe (2003).

    Parameters
    ----------
    fluid_density : float
        Density of fluid (generally, water).
    R: float
        Specific weight..?
    g: float
        Gravitational acceleration.
    mean_active_grain_size: float
        Mean grain size of the 'active' sediment parcels.
    frac_sand: float
        Fraction of the bed surface grain size composed of sand sized parcels.

    Examples
    --------
    >>> from numpy.testing import assert_almost_equal
    >>> assert_almost_equal(
    ...     _calculate_reference_shear_stress(1, 1, 1, 1, 0), 0.036, decimal=2
    ... )
    >>> assert_almost_equal(
    ...     _calculate_reference_shear_stress(1000, 1.65, 9.8, 0.1, 0.9),
    ...     33.957,
    ...     decimal=2,
    ... )
    """

    taursg = (
        fluid_density
        * R
        * g
        * mean_active_grain_size
        * (0.021 + 0.015 * np.exp(-20.0 * frac_sand))
    )

    if np.any(np.asarray(taursg < 0)):
        raise ValueError(
            "NetworkSedimentTransporter: Reference Shields stress is negative"
        )

    return taursg


def _calculate_parcel_volume_post_abrasion(
    starting_volume, travel_distance, abrasion_rate
):
    """Calculate parcel volumes after abrasion, according to Sternberg
    exponential abrasion.

    Parameters
    ----------
    starting_volume : float or array
        Starting volume of each parcel.
    travel_distance: float or array
        Travel distance for each parcel during this timestep, in ___.
    abrasion_rate: float or array
        Volumetric abrasion exponent (1/m).

    Examples
    --------
    >>> import pytest
    >>> _calculate_parcel_volume_post_abrasion(10,100,0.003)
    7.4081822068171785
    >>> _calculate_parcel_volume_post_abrasion(10,300,0.1)
    9.3576229688401746e-13
    >>> with pytest.raises(ValueError):
    ...     _calculate_parcel_volume_post_abrasion(10,300,-3)

    """

    volume = starting_volume * np.exp(travel_distance * (-abrasion_rate))

    if np.any(volume > starting_volume):
        raise ValueError("NST parcel volume *increases* due to abrasion")

    return volume


def _calculate_parcel_grain_diameter_post_abrasion(
    starting_diameter, pre_abrasion_volume, post_abrasion_volume
):
    """Calculate parcel grain diameters after abrasion, according to Sternberg
    exponential abrasion.

    Parameters
    ----------
    starting_diameter : float or array
        Starting volume of each parcel.
    pre_abrasion_volume: float or array
        Parcel volume before abrasion.
    post_abrasion_volume: float or array
        Parcel volume after abrasion.

    Examples
    --------
    >>> import numpy as np
    >>> from numpy.testing import assert_almost_equal

    If no abrasion happens, we should get the same value.

    >>> _calculate_parcel_grain_diameter_post_abrasion(10, 1, 1)
    10.0

    If some abrasion happens, test the value.

    >>> starting_diameter = 10
    >>> pre_abrasion_volume = 2
    >>> post_abrasion_volume = 1
    >>> expected_value = (
    ...     starting_diameter *
    ...     ( post_abrasion_volume / pre_abrasion_volume) ** (1. / 3.))
    >>> print(np.round(expected_value, decimals=3))
    7.937
    >>> assert_almost_equal(
    ...    _calculate_parcel_grain_diameter_post_abrasion(10, 2, 1),
    ...    expected_value)

    """

    abraded_grain_diameter = starting_diameter * (
        post_abrasion_volume / pre_abrasion_volume
    ) ** (1.0 / 3.0)

    return abraded_grain_diameter



# ! /usr/env/python
"""Base class for profile constructors."""

from abc import ABC, abstractmethod

import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection

from landlab import Component
from landlab.plot import imshow_grid
from landlab.utils.return_array import return_array_at_node


def _recursive_max(jagged):
    """
    Examples
    --------
    from landlab.components.profiler.base_profiler import _recursive_max
    >>> struct = [[1, 2, 3, 4],
    ...           [[2, 3, 4, 5],
    ...            [3, 4, 5, 6]],
    ...           [4, 5, 6, 7]]
    >>> _recursive_max(struct)
    7
    >>> _recursive_max([100])
    100
    """
    return max(_recursive_max(j) if hasattr(j, "__iter__") else j for j in jagged)


def _recursive_min(jagged):
    """
    Examples
    --------
    from landlab.components.profiler.base_profiler import _recursive_min
    >>> struct = [[1, 2, 3, 4],
    ...           [[2, 3, 4, 5],
    ...            [3, 4, 5, 6]],
    ...           [4, 5, 6, 7]]
    >>> _recursive_min(struct)
    1
    >>> _recursive_min([100])
    100
    """
    return min(_recursive_min(j) if hasattr(j, "__iter__") else j for j in jagged)


class _BaseProfiler(ABC, Component):
    """Base class to handle profilers.

    Primarily exists to handle plotting.
    """

    _name = "_BaseProfiler"

    _unit_agnostic = True

    _info = {}

    def __init__(self, grid):
        super().__init__(grid)

    def run_one_step(self):
        """Calculate the profile data structure and distances along it."""
        # calculate the profile IDs data structure.
        self._create_profile_structure()

    @abstractmethod
    def _create_profile_structure(self):
        """Private class for creating profile structure.

        Expectation is that this will be overridden to create the following
        three private attributes:

        self._nodes
        self._distance_along_profile

        are each lists of numpy arrays, one array per segment.

        self._colors

        is a list of RGBA tuples, one tuple per segment.

        The order of segments is expected to be consistent between each of the
        three data structures.
        """
        ...  # pragma: no cover

    @property
    def distance_along_profile(self):
        """List of distances along profile for each segment.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import (
        ...     FastscapeEroder,
        ...     FlowAccumulator,
        ...     ChannelProfiler
        ...     )
        >>> mg = RasterModelGrid((10, 10), xy_spacing=10)
        >>> np.random.seed(42)
        >>> z = mg.add_zeros('topographic__elevation', at='node')
        >>> z[mg.core_nodes] += np.random.randn(mg.core_nodes.size)
        >>> fa = FlowAccumulator(mg)
        >>> sp = FastscapeEroder(mg, K_sp=0.0001)
        >>> dt = 1000
        >>> for i in range(200):
        ...     fa.run_one_step()
        ...     sp.run_one_step(dt=dt)
        ...     z[mg.core_nodes] += 0.001 * dt
        >>> profiler = ChannelProfiler(mg)
        >>> profiler.run_one_step()
        >>> profiler.distance_along_profile
        [array([  0.,  10.,  20.,  30.,  40.,  50.])]
        """
        return self._distance_along_profile

    @property
    def nodes(self):
        """List of node ids for each segment.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import (
        ...     FastscapeEroder,
        ...     FlowAccumulator,
        ...     ChannelProfiler
        ...     )
        >>> mg = RasterModelGrid((10, 10), xy_spacing=10)
        >>> np.random.seed(42)
        >>> z = mg.add_zeros('topographic__elevation', at='node')
        >>> z[mg.core_nodes] += np.random.randn(mg.core_nodes.size)
        >>> fa = FlowAccumulator(mg)
        >>> sp = FastscapeEroder(mg, K_sp=0.0001)
        >>> dt = 1000
        >>> for i in range(200):
        ...     fa.run_one_step()
        ...     sp.run_one_step(dt=dt)
        ...     z[mg.core_nodes] += 0.001 * dt
        >>> profiler = ChannelProfiler(mg)
        >>> profiler.run_one_step()
        >>> profiler.nodes
        [array([59, 58, 57, 56, 46, 45])]
        """
        return self._nodes

    @property
    def colors(self):
        """List of colors for each segment.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import (
        ...     FastscapeEroder,
        ...     FlowAccumulator,
        ...     ChannelProfiler
        ...     )
        >>> mg = RasterModelGrid((10, 10), xy_spacing=10)
        >>> np.random.seed(42)
        >>> z = mg.add_zeros('topographic__elevation', at='node')
        >>> z[mg.core_nodes] += np.random.randn(mg.core_nodes.size)
        >>> fa = FlowAccumulator(mg)
        >>> sp = FastscapeEroder(mg, K_sp=0.0001)
        >>> dt = 1000
        >>> for i in range(200):
        ...     fa.run_one_step()
        ...     sp.run_one_step(dt=dt)
        ...     z[mg.core_nodes] += 0.001 * dt
        >>> profiler = ChannelProfiler(mg)
        >>> profiler.run_one_step()
        >>> np.round(profiler.colors, decimals=2)
        array([[ 0.27,  0.  ,  0.33,  1.  ]])
        """
        return self._colors

    def plot_profiles(
        self,
        field="topographic__elevation",
        xlabel="Distance Along Profile",
        ylabel="Plotted Quantity",
        title="Extracted Profiles",
        color=None,
    ):
        """Plot distance-upstream vs at at-node or size (nnodes,) quantity.

        Parameters
        ----------
        field : field name or nnode array
            Array of the at-node-field to plot against distance upstream.
            Default value is the at-node field 'topographic__elevation'.
        xlabel : str, optional
            X-axis label, default is "Distance Along Profile".
        ylabel : str, optional
            Y-axis label, default value is "Plotted Quantity".
        title : str, optional
            Plot title, default value is "Extracted Profiles".
        color : RGBA tuple or color string
            Color to use in order to plot all profiles the same color. Default
            is None, and the colors assigned to each profile are used.
        """
        quantity = return_array_at_node(self._grid, field)

        # create segments the way that line collection likes them.
        segments = []
        qmin = []
        qmax = []
        for idx, nodes in enumerate(self._nodes):
            segments.append(
                list(zip(self._distance_along_profile[idx], quantity[nodes]))
            )
            qmin.append(min(quantity[nodes]))
            qmax.append(max(quantity[nodes]))

        # We need to set the plot limits.
        ax = plt.gca()
        ax.set_xlim(
            _recursive_min(self._distance_along_profile),
            _recursive_max(self._distance_along_profile),
        )
        ax.set_ylim(min(qmin), max(qmax))

        line_segments = LineCollection(segments)
        colors = color or self._colors
        line_segments.set_color(colors)
        ax.add_collection(line_segments)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title(title)

    def plot_profiles_in_map_view(
        self, field="topographic__elevation", endpoints_only=False, color=None, **kwds
    ):
        """Plot profile locations in map view.

        Parameters
        ----------
        field : field name or nnode array
            Array of the at-node-field to plot as the 2D map values.
            Default value is the at-node field 'topographic__elevation'.
        endpoints_only : boolean
            Boolean where False (default) indicates every node along the
            profile is plotted, or True indicating only segment endpoints are
            plotted.
        color : RGBA tuple or color string
            Color to use in order to plot all profiles the same color. Default
            is None, and the colors assigned to each profile are used.
        **kwds : dictionary
            Keyword arguments to pass to imshow_grid.
        """
        # make imshow_grid background
        imshow_grid(self._grid, field, **kwds)
        ax = plt.gca()

        # create segments the way that line collection likes them.
        segments = []
        for nodes in self._nodes:
            if endpoints_only:
                select_nodes = [nodes[0], nodes[-1]]
                segments.append(
                    list(
                        zip(
                            self._grid.x_of_node[select_nodes],
                            self._grid.y_of_node[select_nodes],
                        )
                    )
                )

            else:
                segments.append(
                    list(zip(self._grid.x_of_node[nodes], self._grid.y_of_node[nodes]))
                )

        line_segments = LineCollection(segments)
        colors = color or self._colors
        line_segments.set_color(colors)
        ax.add_collection(line_segments)



import numpy as np

cimport cython
cimport numpy as np

DTYPE_INT = int
ctypedef np.int_t DTYPE_INT_t

DTYPE_FLOAT = np.double
ctypedef np.double_t DTYPE_FLOAT_t


@cython.boundscheck(False)
@cython.wraparound(False)
def get_matrix_diagonal_elements_with_coef(
    np.ndarray[DTYPE_INT_t, ndim=2] core_nodes_at_c2c_link,
    np.ndarray[DTYPE_INT_t, ndim=2] core_nodes_at_c2fv_link,
    np.ndarray[DTYPE_INT_t, ndim=2] core_nodes_at_fv2c_link,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] coef_at_c2c_link,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] coef_at_c2fv_link,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] coef_at_fv2c_link,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] data,
):
    cdef int tail, head
    cdef int link
    cdef int n_links
    cdef double coef

    n_links = len(core_nodes_at_c2c_link)
    for link in range(n_links):
        tail = core_nodes_at_c2c_link[link, 0]
        head = core_nodes_at_c2c_link[link, 1]
        coef = coef_at_c2c_link[link]
        data[tail] -= coef
        data[head] -= coef

    n_links = len(core_nodes_at_c2fv_link)
    for link in range(n_links):
        tail = core_nodes_at_c2fv_link[link, 0]
        data[tail] -= coef_at_c2fv_link[link]

    n_links = len(core_nodes_at_fv2c_link)
    for link in range(n_links):
        head = core_nodes_at_fv2c_link[link, 1]
        data[head] -= coef_at_fv2c_link[link]


@cython.boundscheck(False)
@cython.wraparound(False)
def get_matrix_diagonal_elements(
    np.ndarray[DTYPE_INT_t, ndim=2] core_nodes_at_c2c_link,
    np.ndarray[DTYPE_INT_t, ndim=2] core_nodes_at_c2fv_link,
    np.ndarray[DTYPE_INT_t, ndim=2] core_nodes_at_fv2c_link,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] data,
):
    cdef int tail, head
    cdef int link
    cdef int n_links

    n_links = len(core_nodes_at_c2c_link)
    for link in range(n_links):
        tail = core_nodes_at_c2c_link[link, 0]
        head = core_nodes_at_c2c_link[link, 1]
        data[tail] -= 1.0
        data[head] -= 1.0

    n_links = len(core_nodes_at_c2fv_link)
    for link in range(n_links):
        tail = core_nodes_at_c2fv_link[link, 0]
        data[tail] -= 1.0

    n_links = len(core_nodes_at_fv2c_link)
    for link in range(n_links):
        head = core_nodes_at_fv2c_link[link, 1]
        data[head] -= 1.0


@cython.boundscheck(False)
@cython.wraparound(False)
def fill_right_hand_side(
    np.ndarray[DTYPE_INT_t, ndim=2] nodes_at_c2fv_link,
    np.ndarray[DTYPE_INT_t, ndim=2] nodes_at_fv2c_link,
    np.ndarray[DTYPE_INT_t, ndim=1] core_node_at_node,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] value_at_node,
    np.ndarray[DTYPE_FLOAT_t, ndim=1] out,
):
    cdef int tail, head

    for tail, head in nodes_at_c2fv_link:
        out[core_node_at_node[tail]] -= value_at_node[head]

    for tail, head in nodes_at_fv2c_link:
        out[core_node_at_node[head]] -= value_at_node[tail]



# ! /usr/env/python

from .channel_profiler import ChannelProfiler
from .profiler import Profiler
from .trickle_down_profiler import TrickleDownProfiler

__all__ = ["ChannelProfiler", "Profiler", "TrickleDownProfiler"]



"""Created on Mon Oct 19.

@author: dejh
"""
import contextlib

import numpy as np

from landlab import Component, FieldError, RasterModelGrid
from landlab.components import DepressionFinderAndRouter, FlowAccumulator

# TODO: this should probably follow Barnes et al., 2014 for max efficiency


class SinkFiller(Component):
    """This component identifies depressions in a topographic surface, then
    fills them in in the topography.  No attempt is made to conserve sediment
    mass. User may specify whether the holes should be filled to flat, or with
    a gradient downwards towards the depression outlet. The gradient can be
    spatially variable, and is chosen to not reverse any drainage directions at
    the perimeter of each lake.

    The primary method of this class is 'run_one_step'. 'fill_pits' is a
    synonym.

    Constructor assigns a copy of the grid, and calls the initialize
    method.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator, SinkFiller
    >>> import numpy as np
    >>> lake1 = np.array([34, 35, 36, 44, 45, 46, 54, 55, 56, 65, 74])
    >>> lake2 = np.array([78, 87, 88])
    >>> guard_nodes = np.array([23, 33, 53, 63, 73, 83])
    >>> lake = np.concatenate((lake1, lake2))
    >>> mg = RasterModelGrid((10, 10))
    >>> z = np.ones(100, dtype=float)
    >>> z += mg.node_x  # add a slope
    >>> z[guard_nodes] += 0.001  # forces the flow out of a particular node
    >>> z[lake] = 0.
    >>> field = mg.add_field(
    ...     "topographic__elevation",
    ...     z,
    ...     at="node",
    ...     units="-",
    ...     copy=True,
    ... )
    >>> fr = FlowAccumulator(mg, flow_director='D8')
    >>> fr.run_one_step()
    >>> mg.at_node['flow__sink_flag'][mg.core_nodes].sum()
    14
    >>> hf = SinkFiller(mg, apply_slope=False)
    >>> hf.run_one_step()
    >>> np.allclose(mg.at_node['topographic__elevation'][lake1], 4.)
    True
    >>> np.allclose(mg.at_node['topographic__elevation'][lake2], 7.)
    True

    Now reset and demonstrate the adding of an inclined surface:

    >>> field[:] = z
    >>> hf = SinkFiller(mg, apply_slope=True)
    >>> hf.run_one_step()
    >>> hole1 = np.array([4.00007692, 4.00015385, 4.00023077, 4.00030769,
    ...                   4.00038462, 4.00046154, 4.00053846, 4.00061538,
    ...                   4.00069231, 4.00076923, 4.00084615])
    >>> hole2 = np.array([7.4, 7.2, 7.6])
    >>> np.allclose(mg.at_node['topographic__elevation'][lake1], hole1)
    True
    >>> np.allclose(mg.at_node['topographic__elevation'][lake2], hole2)
    True
    >>> fr.run_one_step()
    >>> mg.at_node['flow__sink_flag'][mg.core_nodes].sum()
    0

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    None Listed

    **Additional References**

    Tucker, G., Lancaster, S., Gasparini, N., Bras, R., Rybarczyk, S. (2001).
    An object-oriented framework for distributed hydrologic and geomorphic
    modeling using triangulated irregular networks. Computers & Geosciences
    27(8), 959-973. https://dx.doi.org/10.1016/s0098-3004(00)00134-5

    """

    _name = "SinkFiller"

    _unit_agnostic = True

    _info = {
        "sediment_fill__depth": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Depth of sediment added at eachnode",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
    }

    def __init__(self, grid, routing="D8", apply_slope=False, fill_slope=1.0e-5):
        """
        Parameters
        ----------
        grid : ModelGrid
            A landlab grid.
        routing : {'D8', 'D4'} (optional)
            If grid is a raster type, controls whether fill connectivity can
            occur on diagonals ('D8', default), or only orthogonally ('D4').
            Has no effect if grid is not a raster.
        apply_slope : bool
            If False (default), leave the top of the filled sink flat. If True,
            apply the slope fill_slope to the top surface to allow subsequent flow
            routing. A test is performed to ensure applying this slope will not
            alter the drainage structure at the edge of the filled region
            (i.e., that we are not accidentally reversing the flow direction
            far from the outlet.)
        fill_slope : float (m/m)
            The slope added to the top surface of filled pits to allow flow
            routing across them, if apply_slope.
        """
        super().__init__(grid)

        if "flow__receiver_node" in grid.at_node and grid.at_node[
            "flow__receiver_node"
        ].size != grid.size("node"):
            raise NotImplementedError(
                "A route-to-multiple flow director has been "
                "run on this grid. The landlab development team has not "
                "verified that SinkFiller is compatible with "
                "route-to-multiple methods. Please open a GitHub Issue "
                "to start this process."
            )

        if routing != "D8":
            assert routing == "D4"
        self._routing = routing
        if isinstance(self._grid, RasterModelGrid) and (routing == "D8"):
            self._D8 = True
            self._num_nbrs = 8
        else:
            self._D8 = False  # useful shorthand for thia test we do a lot
            if isinstance(self._grid, RasterModelGrid):
                self._num_nbrs = 4
        self._fill_slope = fill_slope
        self._apply_slope = apply_slope

        self._elev = self._grid.at_node["topographic__elevation"]
        self._topo_field_name = "topographic__elevation"

        # create the only new output field:
        self._sed_fill_depth = self._grid.add_zeros(
            "node", "sediment_fill__depth", clobber=True
        )

        self._lf = DepressionFinderAndRouter(
            self._grid, routing=self._routing, reroute_flow=True
        )
        self._fr = FlowAccumulator(self._grid, flow_director=self._routing)

    def fill_pits(self):
        """This is a synonym for the main method :func:`run_one_step`."""
        self.run_one_step()

    def run_one_step(self):
        """This is the main method.

        Call it to fill depressions in a starting topography.
        """
        self._original_elev = self._elev.copy()
        # We need this, as we'll have to do ALL this again if we manage
        # to jack the elevs too high in one of the "subsidiary" lakes.
        # We're going to implement the lake_mapper component to do the heavy
        # lifting here, then delete its fields. This means we first need to
        # test if these fields already exist, in which case, we should *not*
        # delete them!
        existing_fields = {}
        spurious_fields = set()
        set_of_outputs = set(self._lf.output_var_names) | set(self._fr.output_var_names)
        with contextlib.suppress(KeyError):
            set_of_outputs.remove(self._topo_field_name)
        for field in set_of_outputs:
            try:
                existing_fields[field] = self._grid.at_node[field].copy()
            except FieldError:  # not there; good!
                spurious_fields.add(field)

        self._fr.run_one_step()
        self._lf.map_depressions()
        # add the depression depths to get up to flat:
        self._elev += self._grid.at_node["depression__depth"]
        # if apply_slope is none, we're now done! But if not...
        if self._apply_slope:
            # new way of doing this - use the upstream structure! Should be
            # both more general and more efficient
            for outlet_node, lake_code in zip(
                self._lf.lake_outlets, self._lf.lake_codes
            ):
                lake_nodes = np.where(self._lf.lake_map == lake_code)[0]
                lake_perim = self._get_lake_ext_margin(lake_nodes)
                perim_elevs = self._elev[lake_perim]
                out_elev = self._elev[outlet_node]
                lowest_elev_perim = perim_elevs[perim_elevs != out_elev].min()
                # note we exclude the outlet node
                elev_increment = (lowest_elev_perim - self._elev[outlet_node]) / (
                    lake_nodes.size + 2.0
                )
                assert elev_increment > 0.0
                all_ordering = self._grid.at_node["flow__upstream_node_order"]
                upstream_order_bool = np.in1d(
                    all_ordering, lake_nodes, assume_unique=True
                )
                lake_upstream_order = all_ordering[upstream_order_bool]
                argsort_lake = np.argsort(lake_upstream_order)
                elevs_to_add = (
                    np.arange(lake_nodes.size, dtype=float) + 1.0
                ) * elev_increment
                sorted_elevs_to_add = elevs_to_add[argsort_lake]
                self._elev[lake_nodes] += sorted_elevs_to_add
        # now put back any fields that were present initially, and wipe the
        # rest:
        for delete_me in spurious_fields:
            if delete_me in self._grid.at_node:
                self._grid.delete_field("node", delete_me)
        for update_me in existing_fields.keys():
            self._grid.at_node[update_me][:] = existing_fields[update_me]
        # fill the output field
        self._sed_fill_depth[:] = self._elev - self._original_elev

    def _add_slopes(self, slope, outlet_node, lake_code):
        """Assuming you have already run the lake_mapper, adds an incline
        towards the outlet to the nodes in the lake."""
        new_elevs = self._elev.copy()
        outlet_coord = (self._grid.node_x[outlet_node], self._grid.node_y[outlet_node])
        lake_nodes = np.where(self._lf.lake_map == lake_code)[0]
        lake_nodes = np.setdiff1d(lake_nodes, self._lake_nodes_treated)
        # lake_ext_margin = self._get_lake_ext_margin(lake_nodes)
        d = self._grid.calc_distances_of_nodes_to_point(
            outlet_coord, node_subset=lake_nodes
        )
        add_vals = slope * d
        new_elevs[lake_nodes] += add_vals
        self._lake_nodes_treated = np.union1d(self._lake_nodes_treated, lake_nodes)
        return new_elevs, lake_nodes

    def _get_lake_ext_margin(self, lake_nodes):
        """Returns the nodes forming the external margin of the lake, honoring
        the *routing* method (D4/D8) if applicable."""
        if self._D8 is True:
            all_poss = np.union1d(
                self._grid.active_adjacent_nodes_at_node[lake_nodes],
                self._grid.diagonal_adjacent_nodes_at_node[lake_nodes],
            )
        else:
            all_poss = np.unique(self._grid.active_adjacent_nodes_at_node[lake_nodes])
        lake_ext_edge = np.setdiff1d(all_poss, lake_nodes)
        return lake_ext_edge[lake_ext_edge != self._grid.BAD_INDEX]

    def _get_lake_int_margin(self, lake_nodes, lake_ext_edge):
        """Returns the nodes forming the internal margin of the lake, honoring
        the *routing* method (D4/D8) if applicable."""
        lee = lake_ext_edge
        if self._D8 is True:
            all_poss_int = np.union1d(
                self._grid.active_adjacent_nodes_at_node[lee],
                self._grid.diagonal_adjacent_nodes_at_node[lee],
            )
        else:
            all_poss_int = np.unique(self._grid.active_adjacent_nodes_at_node[lee])
        lake_int_edge = np.intersect1d(all_poss_int, lake_nodes)
        return lake_int_edge[lake_int_edge != self._grid.BAD_INDEX]

    def _apply_slope_current_lake(self, apply_slope, outlet_node, lake_code, sublake):
        """Wraps the _add_slopes method to allow handling of conditions where
        the drainage structure would be changed or we're dealing with a
        sublake."""
        while 1:
            starting_elevs = self._elev.copy()
            self._elev[:], lake_nodes = self._add_slopes(
                apply_slope, outlet_node, lake_code
            )
            # ext_edge = self._get_lake_ext_margin(lake_nodes)
            if sublake:
                break
            else:
                if not self.drainage_directions_change(
                    lake_nodes, starting_elevs, self._elev
                ):
                    break
                else:
                    # put the elevs back...
                    self._elev[lake_nodes] = starting_elevs[lake_nodes]
                    # the slope was too big. Reduce it.
                    apply_slope *= 0.1
        # if we get here, either sublake, or drainage dirs are stable

    def drainage_directions_change(self, lake_nodes, old_elevs, new_elevs):
        """True if the drainage structure at lake margin changes, False
        otherwise."""
        ext_edge = self._get_lake_ext_margin(lake_nodes)
        if self._D8:
            edge_neighbors = np.hstack(
                (
                    self._grid.active_adjacent_nodes_at_node[ext_edge],
                    self._grid.diagonal_adjacent_nodes_at_node[ext_edge],
                )
            )
        else:
            edge_neighbors = self._grid.active_adjacent_nodes_at_node[ext_edge].copy()
        edge_neighbors[edge_neighbors == self._grid.BAD_INDEX] = -1
        # ^value irrelevant
        old_neighbor_elevs = old_elevs[edge_neighbors]
        new_neighbor_elevs = new_elevs[edge_neighbors]
        # enforce the "don't change drainage direction" condition:
        edge_elevs = old_elevs[ext_edge].reshape((ext_edge.size, 1))
        cond = np.allclose(
            (edge_elevs >= old_neighbor_elevs), (edge_elevs >= new_neighbor_elevs)
        )
        # if True, we're good, the tilting didn't mess with the fr
        return not cond



import matplotlib.colors as colors
import numpy as np
import pytest
from matplotlib.colors import Normalize

from landlab.plot import plot_network_and_parcels

network_norm = Normalize(-1, 6)
parcel_color_norm = Normalize(0, 1)
parcel_color_norm2 = colors.LogNorm(vmin=0.01, vmax=1)
parcel_size_norm = Normalize(-1, 1)
parcel_size_norm2 = colors.LogNorm(vmin=0.01, vmax=1)


link_color_options = [
    {},  # empty dictionary = defaults
    {
        "network_color": "r",  # specify some simple modifications.
        "network_linewidth": 0.7,
        "parcel_alpha": 0,
    },
    {
        "link_attribute": "sediment_total_volume",  # use a link attribute
        "parcel_alpha": 0,
    },
    {
        "link_attribute": "sediment_total_volume",
        "network_cmap": "jet",  # change colormap
        "network_norm": network_norm,  # and normalize
        "link_attribute_title": "Total Sediment Volume",
        "parcel_alpha": 0,
        "network_linewidth": 3,
    },
]

parcel_color_options = [
    {},  # empty dictionary = defaults
    {"parcel_color": "r", "parcel_size": 10},  # specify some simple modifications.
    {
        "parcel_color_attribute": "D",  # use a parcel attribute.
        "parcel_color_norm": parcel_color_norm,
        "parcel_color_attribute_title": "Diameter [m]",
        "parcel_alpha": 1.0,
    },
    {"parcel_color_attribute": "abrasion_rate", "parcel_color_cmap": "bone"},
]

parcel_size_options = [
    {},  # empty dictionary = defaults
    {"parcel_color": "b", "parcel_size": 5},  # specify some simple modifications.
    {
        "parcel_size_attribute": "D",  # use a parcel attribute.
        "parcel_size_norm": parcel_size_norm2,
        "parcel_size_attribute_title": "Diameter [m]",
        "parcel_alpha": 1.0,
    },
    {
        "parcel_size_attribute": "abrasion_rate",  # an
        "parcel_size_min": 10,
        "parcel_size_max": 100,
    },
]


@pytest.mark.parametrize(
    "arg", ["synthetic", pytest.param("methow", marks=pytest.mark.slow)]
)
@pytest.mark.parametrize(
    ("l_opts", "pc_opts", "ps_opts"),
    zip(link_color_options, parcel_color_options, parcel_size_options),
)
def test_link_options(arg, l_opts, pc_opts, ps_opts, request):
    nst = request.getfixturevalue(arg)
    grid = nst.grid
    parcels = nst._parcels
    opts = {**l_opts, **pc_opts, **ps_opts}
    plot_network_and_parcels(grid, parcels, parcel_time_index=0, **opts)


@pytest.mark.parametrize("title", ["A random number", None])
@pytest.mark.parametrize(
    "arg", ["synthetic", pytest.param("methow", marks=pytest.mark.slow)]
)
def test_link_array(arg, title, request):
    nst = request.getfixturevalue(arg)
    grid = nst.grid
    parcels = nst._parcels

    random_link = np.random.randn(grid.size("link"))

    opts = {
        "link_attribute": random_link,  # use an array of size link.
        "network_cmap": "jet",  # change colormap
        "network_norm": network_norm,  # and normalize
        "link_attribute_title": title,
        "parcel_alpha": 0,
        "network_linewidth": 3,
    }
    plot_network_and_parcels(grid, parcels, parcel_time_index=0, **opts)


@pytest.mark.parametrize(
    "arg", ["synthetic", pytest.param("methow", marks=pytest.mark.slow)]
)
def test_with_filter(arg, request):
    nst = request.getfixturevalue(arg)
    grid = nst.grid
    parcels = nst._parcels
    parcel_filter = np.zeros((parcels.dataset.dims["item_id"]), dtype=bool)
    parcel_filter[::10] = True
    plot_network_and_parcels(
        grid,
        parcels,
        parcel_time_index=0,
        parcel_filter=parcel_filter,
        link_attribute="sediment_total_volume",
        network_norm=network_norm,
        parcel_alpha=1.0,
        parcel_size_attribute="D",
        parcel_color_attribute="D",
        parcel_color_norm=parcel_color_norm2,
        parcel_size_norm=parcel_size_norm,
        parcel_size_attribute_title="D",
    )


def test_double_network_color(synthetic):
    grid = synthetic.grid
    parcels = synthetic._parcels
    with pytest.raises(ValueError):
        plot_network_and_parcels(
            grid, parcels, link_attribute="sediment_total_volume", network_color="r"
        )


def test_double_parcel_color(synthetic):
    grid = synthetic.grid
    parcels = synthetic._parcels
    with pytest.raises(ValueError):
        plot_network_and_parcels(
            grid, parcels, parcel_color_attribute="D", parcel_color="r"
        )


def test_double_parcel_size(synthetic):
    grid = synthetic.grid
    parcels = synthetic._parcels
    with pytest.raises(ValueError):
        plot_network_and_parcels(
            grid, parcels, parcel_size_attribute="D", parcel_size=3
        )


def test_categorical_parcel_color(synthetic):
    grid = synthetic.grid
    parcels = synthetic._parcels
    "quartzite"
    with pytest.raises(ValueError):
        plot_network_and_parcels(grid, parcels, parcel_color_attribute="quartzite")


def test_categorical_parcel_size(synthetic):
    grid = synthetic.grid
    parcels = synthetic._parcels
    "quartzite"
    with pytest.raises(ValueError):
        plot_network_and_parcels(grid, parcels, parcel_size_attribute="quartzite")


def test_missing_parcel_color(synthetic):
    grid = synthetic.grid
    parcels = synthetic._parcels
    "quartzite"
    with pytest.raises(ValueError):
        plot_network_and_parcels(grid, parcels, parcel_color_attribute="not_here")


def test_missing_parcel_size(synthetic):
    grid = synthetic.grid
    parcels = synthetic._parcels
    "quartzite"
    with pytest.raises(ValueError):
        plot_network_and_parcels(grid, parcels, parcel_size_attribute="not_here")



r"""
stream_power_smooth_threshold.py: Defines the StreamPowerSmoothThresholdEroder,
which is a version of the FastscapeEroder (derived from it).

StreamPowerSmoothThresholdEroder uses a mathematically smooth threshold
formulation, rather than one with a singularity. The erosion rate is defined as
follows:

$\omega = K A^m S$

$E = \omega - \omega_c \left[ 1 - \exp ( -\omega / \omega_c ) \right]$

Created on Sat Nov 26 08:36:49 2016

@author: gtucker
"""

import numpy as np

from ..depression_finder.lake_mapper import _FLOODED
from .cfuncs import smooth_stream_power_eroder_solver
from .fastscape_stream_power import FastscapeEroder


class StreamPowerSmoothThresholdEroder(FastscapeEroder):
    """Stream erosion component with smooth threshold function.

    Parameters
    ----------
    grid : ModelGrid
        A grid.
    K_sp : float, array, or field name
        K in the stream power equation (units vary with other parameters).
    m_sp : float, optional
        m in the stream power equation (power on drainage area).
    n_sp : float, optional, ~ 0.5<n_sp<4.
        n in the stream power equation (power on slope). NOTE: NOT PRESENTLY
        HONORED BY StreamPowerSmoothThresholdEroder (TODO)
    threshold_sp : float (TODO: array, or field name)
        The threshold stream power.
    discharge_field : float, field name, or array, optional
        Discharge [L^2/T]. The default is to use the grid field
        'drainage_area'. To use custom spatially/temporally varying
        rainfall, use 'water__unit_flux_in' to specify water input to the
        FlowAccumulator and use "surface_water__discharge" for this
        keyword argument.
    erode_flooded_nodes : bool (optional)
        Whether erosion occurs in flooded nodes identified by a
        depression/lake mapper (e.g., DepressionFinderAndRouter). When set
        to false, the field *flood_status_code* must be present on the grid
        (this is created by the DepressionFinderAndRouter). Default True.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> rg = RasterModelGrid((3, 4))
    >>> rg.set_closed_boundaries_at_grid_edges(False, True, True, True)
    >>> z = rg.add_zeros('node', 'topographic__elevation')
    >>> z[5] = 2.0
    >>> z[6] = 1.0
    >>> from landlab.components import FlowAccumulator
    >>> fr = FlowAccumulator(rg, flow_director='D4')
    >>> fr.run_one_step()
    >>> from landlab.components import StreamPowerSmoothThresholdEroder
    >>> sp = StreamPowerSmoothThresholdEroder(rg, K_sp=1.0)
    >>> sp.thresholds
    1.0
    >>> sp.run_one_step(dt=1.0)
    >>> import numpy as np
    >>> np.round(z[5:7], 3)
    array([ 1.646,  0.667])
    >>> z[5] = 2.0
    >>> z[6] = 1.0
    >>> import numpy as np
    >>> q = np.zeros(rg.number_of_nodes) + 0.25
    >>> q[6] = 100.0
    >>> sp = StreamPowerSmoothThresholdEroder(rg, K_sp=1.0, discharge_field=q)
    >>> sp.run_one_step(dt=1.0)
    >>> np.round(z[5:7], 3)
    array([ 1.754,  0.164])

    References
    ----------
    **Required Software Citation(s) Specific to this Component**

    Barnhart, K., Glade, R., Shobe, C., Tucker, G. (2019). Terrainbento 1.0: a
    Python package for multi-model analysis in long-term drainage basin
    evolution. Geoscientific Model Development  12(4), 1267--1297.
    https://dx.doi.org/10.5194/gmd-12-1267-2019

    **Additional References**

    Braun, J., Willett, S. (2013). A very efficient O(n), implicit and parallel
    method to solve the stream power equation governing fluvial incision and
    landscape evolution. Geomorphology  180-181(C), 170-179.
    https://dx.doi.org/10.1016/j.geomorph.2012.10.008

    """

    _name = "StreamPowerSmoothThresholdEroder"

    _unit_agnostic = True

    _cite_as = """
    @article{barnhart2019terrain,
      author = {Barnhart, Katherine R and Glade, Rachel C and Shobe, Charles M
                and Tucker, Gregory E},
      title = {{Terrainbento 1.0: a Python package for multi-model analysis in
                long-term drainage basin evolution}},
      doi = {10.5194/gmd-12-1267-2019},
      pages = {1267---1297},
      number = {4},
      volume = {12},
      journal = {Geoscientific Model Development},
      year = {2019},
    }
    """

    _info = {
        "drainage_area": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m**2",
            "mapping": "node",
            "doc": "Upstream accumulated surface area contributing to the node's discharge",
        },
        "flow__link_to_receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "ID of link downstream of each node, which carries the discharge",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
    }

    def __init__(
        self,
        grid,
        K_sp=None,
        m_sp=0.5,
        n_sp=1.0,
        threshold_sp=1.0,
        discharge_field="drainage_area",
        erode_flooded_nodes=True,
    ):
        """Initialize StreamPowerSmoothThresholdEroder."""
        if "flow__receiver_node" in grid.at_node and grid.at_node[
            "flow__receiver_node"
        ].size != grid.size("node"):
            raise NotImplementedError(
                "A route-to-multiple flow director has been "
                "run on this grid. The landlab development team has not "
                "verified that StreamPowerSmoothThresholdEroder is compatible "
                "with route-to-multiple methods. Please open a GitHub Issue "
                "to start this process."
            )

        if not erode_flooded_nodes and "flood_status_code" not in grid.at_node:
            raise ValueError(
                "In order to not erode flooded nodes another component "
                "must create the field *flood_status_code*. You want to "
                "run a lake mapper/depression finder."
            )

        if n_sp != 1.0:
            raise ValueError(
                "StreamPowerSmoothThresholdEroder currently only " "supports n_sp = 1"
            )

        # Call base-class init
        super().__init__(
            grid,
            K_sp=K_sp,
            m_sp=m_sp,
            n_sp=n_sp,
            threshold_sp=threshold_sp,
            discharge_field=discharge_field,
        )

        # Arrays with parameters for use in implicit solver
        self._gamma = grid.empty(at="node")
        self._delta = grid.empty(at="node")

    @property
    def alpha(self):
        """Erosion term divided by link length.

        Alpha is given as::

            alpha = K A^m dt / L

        where K is the erodibility, A is the drainage area, m is the
        drainage area exponent, dt is the timestep, and L is the link length.
        """
        return self._alpha

    @property
    def gamma(self):
        """Erosion threshold times timestep."""
        return self._gamma

    @property
    def thresholds(self):
        """Erosion thresholds."""
        return self._thresholds

    @property
    def delta(self):
        """Erosion term divided by link length and erosion threshold.

        delta is given as::

            delta = K A^m dt / (L * omega_c)

        where K is the erodibility, A is the drainage area, m is the
        drainage area exponent, dt is the timestep, L is the link length, and
        omega_c is the erosion threshold.
        """
        return self._delta

    def run_one_step(self, dt, runoff_rate=None):
        """Run one forward iteration of duration dt.

        Parameters
        ----------
        dt : float
            Time step size
        runoff_rate : (not used yet)
            (to be added later)

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> rg = RasterModelGrid((3, 3))
        >>> rg.set_closed_boundaries_at_grid_edges(False, True, True, True)
        >>> z = rg.add_zeros('node', 'topographic__elevation')
        >>> z[4] = 1.0
        >>> from landlab.components import FlowAccumulator
        >>> fr = FlowAccumulator(rg, flow_director='D4')
        >>> fr.run_one_step()
        >>> from landlab.components import StreamPowerSmoothThresholdEroder
        >>> sp = StreamPowerSmoothThresholdEroder(rg, K_sp=1.0)
        >>> sp.run_one_step(dt=1.0)
        >>> sp.alpha
        array([ 0.,  0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.])
        >>> sp.gamma
        array([ 0.,  0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.])
        >>> sp.delta
        array([ 0.,  0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.])
        """
        if not self._erode_flooded_nodes:
            flood_status = self._grid.at_node["flood_status_code"]
            flooded_nodes = np.nonzero(flood_status == _FLOODED)[0]
        else:
            flooded_nodes = []

        # Set up needed arrays
        #
        # Get shorthand for elevation field ("z"), and for up-to-downstream
        # ordered array of node IDs ("upstream_order_IDs")
        node_id = np.arange(self._grid.number_of_nodes)
        upstream_order_IDs = self._grid["node"]["flow__upstream_node_order"]
        z = self._grid["node"]["topographic__elevation"]
        flow_receivers = self._grid["node"]["flow__receiver_node"]

        # Get an array of flow-link length for each node that has a defined
        # receiver (i.e., that drains to another node).
        defined_flow_receivers = np.not_equal(
            self._grid["node"]["flow__link_to_receiver_node"], self._grid.BAD_INDEX
        )
        defined_flow_receivers[flow_receivers == node_id] = False

        if flooded_nodes is not None:
            defined_flow_receivers[flooded_nodes] = False
        flow_link_lengths = self._grid.length_of_d8[
            self._grid.at_node["flow__link_to_receiver_node"][defined_flow_receivers]
        ]

        # Set up alpha, beta, delta arrays
        #
        #   First, compute drainage area or discharge raised to the power m.
        np.power(self._A, self._m, out=self._A_to_the_m)

        #   Alpha
        self._alpha[~defined_flow_receivers] = 0.0
        self._alpha[defined_flow_receivers] = (
            self._K[defined_flow_receivers]
            * dt
            * self._A_to_the_m[defined_flow_receivers]
            / flow_link_lengths
        )

        #   Gamma
        self._gamma[~defined_flow_receivers] = 0.0

        #   Delta
        self._delta[~defined_flow_receivers] = 0.0

        if isinstance(self._thresholds, np.ndarray):
            self._gamma[defined_flow_receivers] = (
                dt * self._thresholds[defined_flow_receivers]
            )

            self._delta[defined_flow_receivers] = (
                self._K[defined_flow_receivers]
                * self._A_to_the_m[defined_flow_receivers]
            ) / (self._thresholds[defined_flow_receivers] * flow_link_lengths)

            self._delta[defined_flow_receivers][
                self._thresholds[defined_flow_receivers] == 0.0
            ] = 0.0
        else:
            self._gamma[defined_flow_receivers] = dt * self._thresholds
            if self._thresholds == 0:
                self._delta[defined_flow_receivers] = 0.0
            else:
                self._delta[defined_flow_receivers] = (
                    self._K[defined_flow_receivers]
                    * self._A_to_the_m[defined_flow_receivers]
                ) / (self._thresholds * flow_link_lengths)

        # Iterate over nodes from downstream to upstream, using scipy's
        # 'newton' function to find new elevation at each node in turn.
        smooth_stream_power_eroder_solver(
            upstream_order_IDs, flow_receivers, z, self._alpha, self._gamma, self._delta
        )



import numpy as np

from ..graph import Graph
from .voronoi_to_graph import VoronoiDelaunayToGraph


class DelaunayGraph(Graph):

    """Graph of a voronoi grid.

    Examples
    --------
    >>> from landlab.graph import DelaunayGraph
    """

    def __init__(
        self, node_y_and_x, max_node_spacing=None, sort=False, perimeter_links=None
    ):
        """Create a voronoi grid.

        Parameters
        ----------
        nodes : tuple of array_like
            Coordinates of every node. First *y*, then *x*.

        Examples
        --------
        >>> from landlab.graph import DelaunayGraph
        >>> node_x = [0.0, 1.0, 2.0,
        ...           0.9, 1.9, 2.9]
        >>> node_y = [0, 0, 0,
        ...           2, 2, 2]
        >>> graph = DelaunayGraph((node_y, node_x), sort=True)
        >>> graph.x_of_node
        array([ 0. ,  1. ,  2. ,  0.9,  1.9,  2.9])
        >>> graph.y_of_node
        array([ 0.,  0.,  0.,  2.,  2.,  2.])
        >>> graph.nodes_at_link # doctest: +NORMALIZE_WHITESPACE
        array([[0, 1], [1, 2],
               [0, 3], [1, 3], [1, 4], [2, 4], [2, 5],
               [3, 4], [4, 5]])
        >>> graph.links_at_node # doctest: +NORMALIZE_WHITESPACE
        array([[ 0,  2, -1, -1], [ 1,  4,  3,  0], [ 6,  5,  1, -1],
               [ 7,  2,  3, -1], [ 8,  7,  4,  5], [ 8,  6, -1, -1]])
        >>> graph.links_at_patch # doctest: +NORMALIZE_WHITESPACE
        array([[3, 2, 0], [5, 4, 1], [7, 3, 4], [8, 5, 6]])

        >>> graph.nodes_at_patch # doctest: +NORMALIZE_WHITESPACE
        array([[3, 0, 1], [4, 1, 2], [4, 3, 1], [5, 4, 2]])
        """
        mesh = VoronoiDelaunayToGraph(
            np.vstack((node_y_and_x[1], node_y_and_x[0])).T,
            perimeter_links=perimeter_links,
        )

        Graph.__init__(
            self,
            node_y_and_x,
            links=mesh.nodes_at_link,
            patches=mesh.links_at_patch,
            sort=sort,
        )



"""Define collections of fields that are attached to a *Landlab*
:class:`~landlab.graph.graph.Graph`.
"""
import numpy as np
import xarray as xr

from .errors import FieldError, GroupError


def reshape_for_storage(array, field_size=None):
    """Reshape an array to be stored as a field.

    For reshaping rules, see :func:`~.shape_for_storage`.

    Parameters
    ----------
    array : numpy.ndarray
        The array to be stored.
    field_size : int, optional
        The size of the field.

    Returns
    -------
    ndarray
        The (possibly) reshaped array.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.field.graph_field import reshape_for_storage

    The shape will be such that the first dimension in the field size.

    >>> data = np.arange(6)
    >>> reshape_for_storage(data, 3)
    array([[0, 1],
           [2, 3],
           [4, 5]])
    >>> reshape_for_storage(data, 2)
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> reshape_for_storage(data, 6)
    array([0, 1, 2, 3, 4, 5])

    If the array is already the correct shape, just return that array.

    >>> data = np.arange(6).reshape((2, 3))
    >>> reshape_for_storage(data, 2) is data
    True

    :meta private:
    """
    shape = shape_for_storage(array, field_size)
    if shape == array.shape or array.ndim == 0:
        return array
    else:
        return array.reshape(shape)


def shape_for_storage(array, field_size=None):
    """Return the shape an array will be stored as.

    :meta private:

    Parameters
    ----------
    array : numpy.ndarray
        The array to be stored.
    field_size : int, optional
        The size of the field.

    Returns
    -------
    tuple of int
        The shape the array will be stored as.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.field.graph_field import shape_for_storage

    The shape will be such that the first dimension in the field size.

    >>> data = np.arange(6)
    >>> shape_for_storage(data, 3) == (3, 2)
    True
    >>> shape_for_storage(data, 2) == (2, 3)
    True
    >>> shape_for_storage(data, 6) == (6, )
    True

    If a field size is not given, the array will be stored as a
    flattened array.

    >>> shape_for_storage(data) == (6, )
    True
    >>> data = np.arange(6).reshape((3, 2))
    >>> shape_for_storage(data) == (6, )
    True

    For field sizes of 1, the array is always flattened.

    >>> shape_for_storage(data, 1) == (1, 6)
    True

    For scalar arrays, the field size must be 1.

    >>> data = np.array(1.)
    >>> shape_for_storage(data) == (1, )
    True
    >>> shape_for_storage(data, field_size=1) == (1, )
    True

    If the array cannot be shaped into a storage shape, a ``ValueError``
    is raised.

    >>> data = np.array(1.)
    >>> shape_for_storage(data, field_size=4) # DOCTEST: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError: unable to reshape array to field size
    >>> data = np.arange(6.)
    >>> shape_for_storage(data, field_size=4) # DOCTEST: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError: unable to reshape array to field size
    """
    if field_size is None:
        field_size = array.size

    if array.size % field_size != 0:
        raise ValueError(
            "unable to reshape array to field size ({} != {})".format(
                array.size, field_size
            )
        )

    if field_size == array.size:
        shape = (array.size,)
    else:
        shape = (field_size, array.size // field_size)

    return shape


class FieldDataset(dict):
    """Wrap an xarray.Dataset as a landlab field.

    This is a light wrapping of xarray.Dataset. The main differences
    are that a `FieldDataset` can be created with a size but not
    allocate any memory for data arrays until an array is actually
    needed. The setitem method is also overriden so that when arrays
    are added they are stored reshaped in the landlab style. That
    is, shaped as `(n_elements, values_per_element)`.

    :meta private:

    Examples
    --------
    >>> from landlab.field.graph_field import FieldDataset

    >>> ds = FieldDataset("node")
    >>> ds.size is None
    True
    >>> ds.set_value("air_temperature", [1.0, 1.0, 1.0, 1.0])
    >>> ds["air_temperature"]
    array([ 1.,  1.,  1.,  1.])
    >>> ds.size
    4
    >>> ds.set_value("air_temperature", [1.0, 1.0, 1.0])
    Traceback (most recent call last):
    ValueError: unable to reshape array to field size (3 != 4)

    >>> ds = FieldDataset("node", fixed_size=False)
    >>> ds.size is None
    True
    >>> ds.set_value("air_temperature", [1.0, 1.0, 1.0, 1.0])
    >>> ds.size
    4
    >>> ds.set_value("air_temperature", [0.0, 0.0])
    >>> ds.size
    2
    >>> ds["air_temperature"]
    array([ 0.,  0.])
    """

    def __init__(self, name, size=None, fixed_size=True):
        """Create a container to hold a collection of *Landlab* fields.

        Parameters
        ----------
        name : str
            The name of the group that identifies the dataset of fields.
        size : int, optional
            The size of fields that will be held within the dataset.
            If not provided, the size will be set later.
        fixed_size : bool, optional
            The size of fields held within the dataset cannot change.
        """
        self._name = name
        self._size = None
        self._fixed_size = bool(fixed_size)
        self._ds = xr.Dataset()
        self._units = {}

        self.size = size

        super().__init__()

    @property
    def size(self):
        """Size of the field dataset as number of elements.

        Examples
        --------
        >>> from landlab.field.graph_field import FieldDataset

        >>> ds = FieldDataset("grid", size=1)
        >>> ds.set_value("air_temperature", [1.0, 1.0, 1.0])
        >>> ds.set_value("ground_temperature", [0.0, 0.0])
        >>> ds["ground_temperature"]
        array([[ 0.,  0.]])

        >>> ds = FieldDataset("grid", size=1)
        >>> ds.set_value("air_temperature", 0.1)
        >>> ds["air_temperature"]
        array(0.1)
        """
        return self._size

    @size.setter
    def size(self, size):
        if self._size != size:
            if self._size is not None and self.fixed_size:
                raise ValueError(
                    "size has already been set ({size}) and fixed_size is True".format(
                        size=self._size
                    )
                )
            elif not isinstance(size, int) or size < 0:
                raise ValueError(f"size must be a positive integer or None ({size})")
            self._size = size

    @property
    def fixed_size(self):
        """Flag that indicates if arrays added to the dataset must be of a fixed size.

        Examples
        --------
        >>> from landlab.field.graph_field import FieldDataset

        >>> ds = FieldDataset("node", fixed_size=False)
        >>> ds.set_value("air_temperature", [1.0, 1.0, 1.0, 1.0])
        >>> ds.set_value("air_temperature", [0.0, 0.0])
        >>> ds["air_temperature"]
        array([ 0.,  0.])

        >>> ds.fixed_size = True
        >>> ds.size
        2
        >>> ds.set_value("air_temperature", [1.0, 1.0, 1.0])
        Traceback (most recent call last):
        ValueError: unable to reshape array to field size (4 != 2)
        """
        return self._fixed_size

    @fixed_size.setter
    def fixed_size(self, fixed_size):
        self._fixed_size = bool(fixed_size)
        if self._fixed_size:
            self.size = self._ds.dims[self._name]

    @property
    def units(self):
        """Return units for each of the fields of the dataset."""
        return self._units

    def set_units(self, name, new_units):
        """Assign units to a field.

        Parameters
        ----------
        name : str
            The name of the field.
        new_units : str
            The new units of the field.
        """
        self._units[name] = self._ds[name].attrs["units"] = new_units

    @property
    def dataset(self):
        """Return the dataset as an :class:`xarray.Dataset`."""
        return self._ds

    def keys(self):
        """Return the name of the fields held in the dataset."""
        return list(self._ds.variables)

    def set_value(self, name, value_array, attrs=None):
        """Assign values to a field.

        Parameters
        ----------
        name : str
            Name of the field in the dataset.
        value_array : numpy.ndarray
            Array of values to assign to the field.
        attrs : dict, optional
            A `dict` holding attributes to save with the field.
        """
        attrs = attrs or {}
        attrs.setdefault("units", "?")

        value_array = np.asarray(value_array)

        if name in self._ds and self._ds[name].values is value_array:
            self._ds[name].values.shape = shape_for_storage(value_array, self.size)
            return

        if self.fixed_size:
            value_array = reshape_for_storage(value_array, self.size)
        else:
            value_array = reshape_for_storage(value_array, None)

        if value_array.ndim > 0:
            self.size = value_array.shape[0]
        else:
            self.size = value_array.size

        if value_array.ndim == 0:
            dims = ()
        elif value_array.ndim == 1:
            dims = (self._name,)
        else:
            dims = (self._name, name + "_per_" + self._name)

        if name in self._ds:
            self._ds = self._ds.drop_vars(name)

        self._ds.update({name: xr.DataArray(value_array, dims=dims, attrs=attrs)})
        self._units[name] = attrs["units"]

    def pop(self, name):
        """Remove a field from the dataset.

        Parameters
        ----------
        name : str
            The name of the field to remove.

        Returns
        -------
        numpy.ndarray
            The values of the field that was removed.
        """
        array = self._ds[name].values
        self._ds = self._ds.drop_vars(name)
        return array

    def __getitem__(self, name):
        """Get a field's values by name."""
        if isinstance(name, str):
            try:
                return self._ds[name].values
            except KeyError as exc:
                raise FieldError(name) from exc
        else:
            raise TypeError("field name not a string")

    def __setitem__(self, name, value_array):
        """Assign values to a field by name."""
        self.set_value(name, value_array)

    def __contains__(self, name):
        """Return if the dataset contains a named field.

        Parameters
        ----------
        name : str
            A field name.

        Returns
        -------
        bool
            `True` if the dataset contains the field, otherwise `False`.
        """
        return name in self._ds

    def __str__(self):
        """Return a human-readable string of the dataset."""
        return str(self._ds)

    def __repr__(self):
        """Return a string representation of the dataset."""
        return "FieldDataset({name}, size={size}, fixed_size={fixed_size})".format(
            name=repr(self._name),
            size=repr(self.size),
            fixed_size=repr(self.fixed_size),
        )

    def __len__(self):
        """Return the number of fields held in the dataset."""
        return len(self._ds.variables)

    def __iter__(self):
        """Iterate through the fields in the dataset."""
        return iter(self._ds.variables)


class GraphFields:
    """Collection of grouped data-fields.

    The :class:`GraphFields` class holds a set of data fields that are
    separated into *groups*. A typical use for this class would be to define
    the groups as being locations on a grid where the values are defined.
    For instance, the groups could be *node*, *cell*, *link*, and *face*.

    Examples
    --------
    Create two groups of data fields defined at *node* and *cell*. Each set can
    have a different number of values.

    >>> from landlab.field import GraphFields
    >>> fields = GraphFields()
    >>> fields.new_field_location("node", 12)
    >>> fields.new_field_location("cell", 2)

    Create some new value arrays for each of the data fields.

    >>> fields.ones("node")
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])
    >>> fields.zeros("cell")
    array([ 0.,  0.])

    Create new value arrays and add them to the data fields. Because the data
    fields are in different groups (node and cell), they can have the same
    name.

    >>> fields.add_ones("topographic__elevation", at="node")
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])
    >>> fields.at_node["topographic__elevation"]
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])

    >>> fields.add_ones("topographic__elevation", at="cell")
    array([ 1.,  1.])
    >>> fields.at_cell["topographic__elevation"]
    array([ 1.,  1.])

    Each group acts as a :class:`dict` so, for instance, to get the variables names
    in a group use the :meth:`keys` method,

    >>> list(fields.at_cell.keys())
    ['topographic__elevation']

    If the size of the new field location is ``None``, the field will be
    unsized. This means that fields added to this location can be of any
    size.

    >>> fields = GraphFields()
    >>> fields.new_field_location("grid", None)
    >>> fields.at_grid["g"] = 9.81
    >>> fields.at_grid["g"]
    array(9.81)
    >>> fields.at_grid["w"] = (3., 4.)
    >>> fields.at_grid["w"]
    array([ 3.,  4.])

    The dimensions of groups can also be specified when the object is
    instantiated. In this case, group sizes are specified as a dictionary
    with keys being group names and values group sizes.

    >>> fields = GraphFields({"node": 6, "grid": None})
    >>> fields.at_grid["g"] = 9.81
    >>> fields.at_node["x"] = [0, 1, 2, 3, 4, 5]
    >>> fields.at_grid["g"]
    array(9.81)
    >>> fields.at_node["x"]
    array([0, 1, 2, 3, 4, 5])
    """

    def __init__(self, *args, **kwds):
        """Create a new collection of field groups."""
        try:
            dims = args[0]
        except IndexError:
            dims = {}

        self._groups = set()
        for loc in dims:
            self.new_field_location(loc, dims[loc])

        self.default_group = kwds.get("default_group", None)

    def __getitem__(self, name):
        """Get the collection of fields from the named group."""
        try:
            return getattr(self, "at_" + name)
        except AttributeError as exc:
            raise GroupError(name) from exc

    @property
    def default_group(self):
        """Return the name of the group into which fields are put by default."""
        return self._default_group

    @default_group.setter
    def default_group(self, loc):
        if self.has_group(loc) or loc is None:
            self._default_group = loc
        else:
            groups = ", ".join([repr(name) for name in sorted(self._groups)])
            raise ValueError(f"{loc!r}: Group does not exists. Not one of {groups}.")

    def new_field_location(self, loc, size=None):
        """Add a new quantity to a field.

        Create an empty group into which new fields can be added. The new group
        is created but no memory allocated yet. The dictionary of the new group
        can be through a new *at_* attribute of the class instance.

        Parameters
        ----------
        loc: str
            Name of the new group to add to the field.
        size: int, optional
            Number of elements in the new quantity. If not provided, the
            size is set to be the size of the first field added to the group.

        Raises
        ------
        ValueError
            If the field already contains the group.

        Examples
        --------
        Create a collection of fields and add two groups, *node* and *cell*,
        to it.

        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location("node", 12)
        >>> fields.new_field_location("cell", 2)

        The group names in the collection are retrieved with the *groups*
        attribute as a :class:`set`.

        >>> names = list(fields.groups)
        >>> names.sort()
        >>> names
        ['cell', 'node']

        Access the new (empty) groups with the *at_* attributes.

        >>> fields.at_cell
        FieldDataset('cell', size=2, fixed_size=True)
        >>> fields.at_node
        FieldDataset('node', size=12, fixed_size=True)

        >>> fields.new_field_location("core_node")
        >>> fields.at_core_node.size is None
        True
        >>> fields.at_core_node["air__temperature"] = [0, 1]
        >>> fields.at_core_node.size
        2

        :meta landlab: field-add
        """
        dataset_name = "at_" + loc
        if loc not in self._groups:
            setattr(
                self, dataset_name, FieldDataset(loc, size, fixed_size=size is not None)
            )
            self._groups.add(loc)
        else:
            raise ValueError(f"{loc!r} location already exists")

    @property
    def groups(self):
        """List of group names.

        Returns
        -------
        set
            Names of field groupings.
        """
        return self._groups

    def has_group(self, name):
        """Check if a group exists.

        Parameters
        ----------
        name: str
            Name of the group.

        Returns
        -------
        bool
            ``True`` if the field contains *group*, otherwise ``False``.

        Examples
        --------
        Check if the field has the groups named *node* or *cell*.

        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location('node', 12)
        >>> fields.has_group('node')
        True
        >>> fields.has_group('cell')
        False

        :meta landlab: info-field
        """
        return name in self._groups

    def has_field(self, *args, **kwds):
        """has_field(field, at=None)

        Check if a field is in a group.

        Parameters
        ----------
        field: str
            Name of the field.
        at: str, optional
            Name of the group.

        Returns
        -------
        bool
            ``True`` if the group contains the field, otherwise ``False``.

        Examples
        --------
        Check if the field named ``topographic__elevation`` is contained
        in a group.

        >>> from landlab.field import GraphFields

        >>> fields = GraphFields()
        >>> fields.new_field_location("node", 12)
        >>> _ = fields.add_ones("topographic__elevation", at="node")
        >>> fields.has_field("topographic__elevation", at="node")
        True
        >>> fields.has_field("topographic__elevation", at="cell")
        False

        >>> fields = GraphFields()
        >>> fields.new_field_location("node", 12)
        >>> _ = fields.add_ones("topographic__elevation", at="node")
        >>> fields.has_field("node", "topographic__elevation")
        True
        >>> fields.has_field("cell", "topographic__elevation")
        False

        :meta landlab: info-field
        """
        if len(args) == 2:
            group, field = args
        elif len(args) == 1:
            group, field = kwds.pop("at", self.default_group), args[0]
        else:
            raise ValueError("number of arguments must be 1 or 2")
        if group is None:
            raise ValueError("no group provided")

        try:
            return field in self[group]
        except KeyError:
            return False

    def keys(self, group):
        """Return the field names in a group.

        Parameters
        ----------
        group : str
            Group name.

        Returns
        -------
        list
            Names of fields held in the given group.

        Examples
        --------
        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location("node", 4)
        >>> list(fields.keys("node"))
        []
        >>> _ = fields.add_empty("topographic__elevation", at="node")
        >>> list(fields.keys("node"))
        ['topographic__elevation']

        :meta landlab: info-field
        """
        return self[group].keys()

    def size(self, group):
        """Return the size of the arrays stored in a group.

        Parameters
        ----------
        group : str
            Group name.

        Returns
        -------
        int
            Array size.

        Examples
        --------
        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location("node", 4)
        >>> fields.size("node")
        4

        :meta landlab: info-grid, info-field
        """
        return self[group].size

    def field_values(self, *args, **kwds):
        """field_values(field, at=None)

        Return the values of a field.

        Given a *group* and a *field*, return a reference to the associated
        data array.

        Parameters
        ----------
        field: str
            Name of the field within *group*.
        at: str, optional
            Name of the group.

        Returns
        -------
        array
            The values of the field.

        Raises
        ------
        landlab.field.errors.GroupError
            If *group* does not exist
        landlab.field.errors.FieldError
            If *field* does not exist

        Examples
        --------
        Create a group of fields called *node*.

        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location("node", 4)

        Add a field, initialized to ones, called *topographic__elevation*
        to the *node* group. The *field_values* method returns a reference
        to the field's data.

        >>> _ = fields.add_ones("topographic__elevation", at="node")
        >>> fields.field_values("topographic__elevation", at="node")
        array([ 1.,  1.,  1.,  1.])

        Raise FieldError if *field* does not exist in *group*.

        >>> fields.field_values("planet_surface__temperature", at="node")
        ...     # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        FieldError: planet_surface__temperature

        If *group* does not exists, raise :class:`~landlab.field.errors.GroupError`.

        >>> fields.field_values("topographic__elevation", at="cell")
        ...     # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        GroupError: cell

        :meta landlab: field-io
        """
        if len(args) == 2:
            group, field = args
        elif len(args) == 1:
            group, field = kwds.pop("at", self.default_group), args[0]
        else:
            raise ValueError("number of arguments must be 1 or 2")
        if group is None:
            raise ValueError("no group provided")

        try:
            fields = self[group]
        except KeyError as exc:
            groups = ", ".join([repr(g) for g in sorted(self._groups)])
            raise GroupError(f"{group!r}: Not one of {groups}.") from exc
        try:
            return fields[field]
        except KeyError as exc:
            raise FieldError(f"{field!r}") from exc

    def return_array_or_field_values(self, *args, **kwds):
        """return_array_or_field_values(field, at=None)

        Return field given a field name, or array of with the correct shape.

        Given a *group* and a *field*, return a reference to the associated
        data array. *field* is either a string that is a field in the group
        or an array of the correct size.

        This function is meant to serve like the
        :class:`~landlab.utils.decorators.use_field_name_or_array` decorator for
        bound functions.

        Parameters
        ----------
        field: str or array
            Name of the field withing *group*.
        at: str, optional
            Name of the group.

        Returns
        -------
        numpy.ndarray
            The values of the field.

        Raises
        ------
        landlab.field.errors.GroupError
            If *group* does not exist
        landlab.field.errors.FieldError
            If *field* does not exist

        Examples
        --------
        Create a group of fields called *node*.

        >>> import numpy as np
        >>> from landlab.field import GraphFields
        >>> fields = GraphFields()
        >>> fields.new_field_location('node', 4)

        Add a field, initialized to ones, called *topographic__elevation*
        to the *node* group. The *field_values* method returns a reference
        to the field's data.

        >>> _ = fields.add_ones("topographic__elevation", at="node")
        >>> fields.field_values("topographic__elevation", at="node")
        array([ 1.,  1.,  1.,  1.])

        Alternatively, if the second argument is an array, its size is
        checked and returned if correct.

        >>> vals = np.array([4., 5., 7., 3.])
        >>> fields.return_array_or_field_values(vals, at="node")
        array([ 4.,  5.,  7.,  3.])

        Raise FieldError if *field* does not exist in *group*.

        >>> fields.return_array_or_field_values("surface__temperature", at="node")
        ...     # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        FieldError: surface__temperature

        If *group* does not exists, Raise GroupError.

        >>> fields.return_array_or_field_values("topographic__elevation", at="cell")
        ...     # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        GroupError: cell

        And if the array of values provided is incorrect, raise a :class:`ValueError`.

        >>> vals = np.array([3., 2., 1.])
        >>> fields.return_array_or_field_values(vals, at="node")
        ...     # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ValueError: Array has incorrect size.

        :meta landlab: field-io
        """
        if len(args) == 2:
            group, field = args
        elif len(args) == 1:
            group, field = kwds.pop("at", self.default_group), args[0]
        else:
            raise ValueError("number of arguments must be 1 or 2")
        if group is None:
            raise ValueError("no group provided")

        if isinstance(field, str):
            vals = self.field_values(group, field)
        else:
            vals = np.asarray(field)
            if vals.size != self[group].size:
                raise ValueError(
                    f"Incorrect array size. The array size of {vals.size} does not "
                    "match that of the group, {group!r}, which has a size of "
                    f"{self[group].size}."
                )
        return vals

    def field_units(self, *args, **kwds):
        """field_units(field, at=None)

        Get units for a field.

        Returns the unit string associated with the data array in *group* and
        *field*.

        Parameters
        ----------
        field: str
            Name of the field withing *group*.
        at: str, optional
            Name of the group.

        Returns
        -------
        str
            The units of the field.

        Raises
        ------
        KeyError
            If either *field* or *group* does not exist.


        :meta landlab: info-field
        """
        if len(args) == 2:
            group, field = args
        elif len(args) == 1:
            group, field = kwds.pop("at", self.default_group), args[0]
        else:
            raise ValueError("number of arguments must be 1 or 2")
        if group is None:
            raise ValueError("no group provided")

        return self[group]._ds[field].attrs["units"]

    def empty(self, *args, **kwds):
        """Uninitialized array whose size is that of the field.

        Return a new array of the data field size, without initializing
        entries. Keyword arguments are the same as that for the equivalent
        *numpy* function.

        Parameters
        ----------
        group : str
            Name of the group.

        See Also
        --------
        numpy.empty : See for a description of optional keywords.
        GraphFields.ones : Equivalent method that initializes the data to 1.
        GraphFields.zeros : Equivalent method that initializes the data to 0.

        Examples
        --------
        >>> from landlab.field import GraphFields
        >>> field = GraphFields()
        >>> field.new_field_location("node", 4)
        >>> field.empty("node") # doctest: +SKIP
        array([  2.31584178e+077,  -2.68156175e+154,   9.88131292e-324,
        ... 2.78134232e-309]) # Uninitialized memory

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys("node"))
        []

        :meta landlab: field-add
        """
        if len(args) == 0:
            group = kwds.pop("at", kwds.pop("centering", "node"))
        else:
            group = args[0]

        if group == "grid":
            raise ValueError(
                "ones is not supported for at='grid', if you "
                "want to create a field at the grid, use\n"
                "grid.at_grid['value_name']=value\n"
                "instead.\nAlternatively, if you want ones"
                "of the shape stored at_grid, use np.array(1)."
            )

        size = getattr(self, f"at_{group}").size
        if size is None:
            raise ValueError("{group!r}: Group is not yet sized.")

        return np.empty(size, **kwds)

    def ones(self, *args, **kwds):
        """Array, initialized to 1, whose size is that of the field.

        Return a new array of the data field size, filled with ones. Keyword
        arguments are the same as that for the equivalent *numpy* function.

        Parameters
        ----------
        group : str
            Name of the group.

        See Also
        --------
        numpy.ones : See for a description of optional keywords.
        GraphFields.empty : Equivalent method that does not initialize the new array.
        GraphFields.zeros : Equivalent method that initializes the data to 0.

        Examples
        --------
        >>> from landlab.field import GraphFields
        >>> field = GraphFields()
        >>> field.new_field_location("node", 4)
        >>> field.ones("node")
        array([ 1.,  1.,  1.,  1.])
        >>> field.ones("node", dtype=int)
        array([1, 1, 1, 1])

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys("node"))
        []

        :meta landlab: field-add
        """
        allocated = self.empty(*args, **kwds)
        allocated.fill(1)
        return allocated

    def zeros(self, *args, **kwds):
        """Array, initialized to 0, whose size is that of the field.

        Parameters
        ----------
        group : str
            Name of the group.

        Return a new array of the data field size, filled with zeros. Keyword
        arguments are the same as that for the equivalent *numpy* function.

        This method is not valid for the group *grid*.

        See Also
        --------
        numpy.zeros : See for a description of optional keywords.
        GraphFields.empty : Equivalent method that does not initialize the new array.
        GraphFields.ones : Equivalent method that initializes the data to 1.

        Examples
        --------
        >>> from landlab.field import GraphFields
        >>> field = GraphFields()
        >>> field.new_field_location("node", 4)
        >>> field.zeros("node")
        array([ 0.,  0.,  0.,  0.])

        Note that a new field is *not* added to the collection of fields.

        >>> list(field.keys("node"))
        []

        :meta landlab: field-add
        """
        allocated = self.empty(*args, **kwds)
        allocated.fill(0)
        return allocated

    def add_field(self, *args, **kwds):
        """add_field(name, value_array, at='node', units='-', copy=False, clobber=False)

        Add an array of values to the field.

        Add an array of data values to a collection of fields and associate it
        with the key, *name*. Use the *copy* keyword to, optionally, add a
        copy of the provided array.

        In the case of adding to the collection *grid*, the added field is a
        *numpy* scalar rather than a *numpy* array.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        value_array : numpy.array
            Array of values to add to the field.
        at : str, optional
            Grid location to store values. If not given, values are
            assumed to be on `node`.
        units : str, optional
            Optionally specify the units of the field.
        copy : bool, optional
            If True, add a *copy* of the array to the field. Otherwise save add
            a reference to the array.
        clobber : bool, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        numpy.ndarray
            The data array added to the field. Depending on the *copy*
            keyword, this could be a copy of *value_array* or *value_array*
            itself.

        Raises
        ------
        ValueError
            If *value_array* has a size different from the field.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.field import GraphFields
        >>> field = GraphFields()
        >>> field.new_field_location("node", 4)
        >>> values = np.ones(4, dtype=int)
        >>> field.add_field("topographic__elevation", values, at="node")
        array([1, 1, 1, 1])

        A new field is added to the collection of fields. The saved value
        array is the same as the one initially created.

        >>> field.at_node["topographic__elevation"] is values
        True

        If you want to save a copy of the array, use the *copy* keyword. In
        addition, adding values to an existing field will remove the reference
        to the previously saved array. The *clobber=False* keyword changes this
        behavior to raise an exception in such a case.

        >>> field.add_field(
        ...     "topographic__elevation", values, at="node", copy=True, clobber=True
        ... )
        array([1, 1, 1, 1])
        >>> field.at_node["topographic__elevation"] is values
        False
        >>> field.add_field(
        ...     "topographic__elevation", values, at="node", clobber=False
        ... ) # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        FieldError: topographic__elevation

        :meta landlab: field-add
        """
        if len(args) == 3:
            at, name, value_array = args
        elif len(args) == 2:
            at, name, value_array = (kwds.pop("at", None), args[0], args[1])
        else:
            raise ValueError("number of arguments must be 2 or 3")

        units = kwds.get("units", "?")
        copy = kwds.get("copy", False)
        clobber = kwds.get("clobber", False)
        value_array = np.asarray(value_array)

        at = at or self.default_group
        if at is None:
            raise ValueError("no group specified")

        attrs = {"long_name": name, "units": units}

        if copy:
            value_array = value_array.copy()

        ds = getattr(self, "at_" + at)

        if not clobber and name in ds:
            raise FieldError(
                f"Unable to add the field, {name!r}, to the group, {at!r}, because a field "
                "with that name already exists in that group. "
                "Use `clobber=True` to replace the existing field. "
                f"For example, grid.add_field({name!r}, at={at!r}, clobber=True)"
            )

        dims = (at,)
        if value_array.ndim > 1:
            dims += (name + "_per_" + at,)
            value_array = value_array.reshape((value_array.shape[0], -1))

        ds.set_value(name, value_array, attrs=attrs)
        # ds[name] = value_array
        return ds[name]

    def delete_field(self, loc, name):
        """Erases an existing field.

        Parameters
        ----------
        loc : str
            Name of the group.
        name: str
            Name of the field.

        Raises
        ------
        KeyError
            If the named field does not exist.


        :meta landlab: field-add
        """
        try:
            ds = getattr(self, "at_" + loc)
        except AttributeError as exc:
            raise KeyError(loc) from exc
        ds._ds = ds._ds.drop_vars(name)

    def add_empty(self, *args, **kwds):
        """add_empty(name, at='node', units='-', clobber=False)

        Create and add an uninitialized array of values to the field.

        Create a new array of the data field size, without initializing
        entries, and add it to the field as *name*. The *units* keyword gives
        the units of the new fields as a string. Remaining keyword arguments
        are the same as that for the equivalent *numpy* function.

        This method is not valid for the group *grid*.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        at : str, optional
            Grid location to store values. If not given, values are
            assumed to be on `node`.
        units : str, optional
            Optionally specify the units of the field.
        clobber : bool, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        numpy.ndarray
            A reference to the newly-created array.

        See Also
        --------
        numpy.empty : See for a description of optional keywords.
        GraphFields.empty : Equivalent method that does not initialize the new array.
        GraphFields.zeros : Equivalent method that initializes the data to 0.


        :meta landlab: field-add
        """
        if len(args) == 2:
            loc, name = args
        elif len(args) == 1:
            loc, name = kwds.pop("at"), args[0]
        else:
            raise ValueError("number of arguments must be 1 or 2")
        units = kwds.pop("units", "?")
        copy = kwds.pop("copy", False)
        clobber = kwds.pop("clobber", False)
        return self.add_field(
            name,
            self.empty(at=loc, **kwds),
            at=loc,
            units=units,
            copy=copy,
            clobber=clobber,
        )

    def add_ones(self, *args, **kwds):
        """add_ones(name, at='node', units='-', clobber=False)

        Create and add an array of values, initialized to 1, to the field.

        Create a new array of the data field size, filled with ones, and
        add it to the field as *name*. The *units* keyword gives the units of
        the new fields as a string. Remaining keyword arguments are the same
        as that for the equivalent *numpy* function.

        This method is not valid for the group *grid*.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        at : str, optional
            Grid location to store values. If not given, values are
            assumed to be on `node`.
        units : str, optional
            Optionally specify the units of the field.
        clobber : bool, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        numpy.ndarray
            A reference to the newly-created array.

        See Also
        --------
        numpy.ones : See for a description of optional keywords.
        GraphFields.add_empty : Equivalent method that does not initialize the new array.
        GraphFields.add_zeros : Equivalent method that initializes the data to 0.

        Examples
        --------
        Add a new, named field to a collection of fields.

        >>> from landlab.field import GraphFields
        >>> field = GraphFields()
        >>> field.new_field_location('node', 4)
        >>> field.add_ones("topographic__elevation", at="node")
        array([ 1.,  1.,  1.,  1.])
        >>> list(field.keys('node'))
        ['topographic__elevation']
        >>> field['node']['topographic__elevation']
        array([ 1.,  1.,  1.,  1.])
        >>> field.at_node['topographic__elevation']
        array([ 1.,  1.,  1.,  1.])

        :meta landlab: field-add
        """
        data = self.add_empty(*args, **kwds)
        data.fill(1)
        return data

    def add_zeros(self, *args, **kwds):
        """add_zeros(name, at='node', units='-', clobber=False)

        Create and add an array of values, initialized to 0, to the field.

        Create a new array of the data field size, filled with zeros, and
        add it to the field as *name*. The *units* keyword gives the units of
        the new fields as a string. Remaining keyword arguments are the same
        as that for the equivalent *numpy* function.

        :meta landlab: field-add

        Parameters
        ----------
        name : str
            Name of the new field to add.
        at : str, optional
            Grid location to store values. If not given, values are
            assumed to be on `node`.
        units : str, optional
            Optionally specify the units of the field.
        clobber : bool, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        array :
            A reference to the newly-created array.

        See also
        --------
        numpy.zeros : See for a description of optional keywords.
        GraphFields.add_empty : Equivalent method that does not initialize the new array.
        GraphFields.add_ones : Equivalent method that initializes the data to 1.


        :meta landlab: field-add
        """
        data = self.add_empty(*args, **kwds)
        data.fill(0)
        return data

    def add_full(self, *args, **kwds):
        """Create and add an array of values, fill with `fill_value`.

        Parameters
        ----------
        name : str
            Name of the new field to add.
        fill_value : scalar
            Fill value.
        at : str, optional
            Grid location to store values. If not given, values are
            assumed to be on `node`.
        units : str, optional
            Optionally specify the units of the field.
        copy : bool, optional
            If True, add a *copy* of the array to the field. Otherwise save add
            a reference to the array.
        clobber : bool, optional
            Raise an exception if adding to an already existing field.

        Returns
        -------
        numpy.ndarray
            A reference to the newly-created array.


        :meta landlab: field-add
        """
        if len(args) == 3:
            at, name, fill_value = args
        elif len(args) == 2:
            at = kwds.pop("at", "node")
            name, fill_value = args
        else:
            raise ValueError("number of arguments must be 2 or 3")

        data = self.add_empty(name, at=at, **kwds)
        data.fill(fill_value)
        return data



#!/usr/bin/env python
"""Tests for SpeciesEvolver zone objects."""
import numpy as np
import pandas as pd
import pytest

from landlab import RasterModelGrid
from landlab.components import SpeciesEvolver
from landlab.components.species_evolution import ZoneController, ZoneTaxon, zone as zn


@pytest.fixture()
def zone_example_grid():
    mg = RasterModelGrid((5, 7), 2)
    z = mg.add_zeros("node", "topographic__elevation")
    return mg, z


def zone_func(grid):
    z = grid.at_node["topographic__elevation"]
    return z == 1


def zone_func_with_vars(grid, var1, var2):
    z = grid.at_node["topographic__elevation"]
    return np.all([z == 1, grid.x_of_node > var1, grid.y_of_node > var2], 0)


def test_none_to_none(zone_example_grid):
    mg, z = zone_example_grid

    sc = ZoneController(mg, zone_func)
    sc.run_one_step(1)

    np.testing.assert_array_equal(len(sc.zones), 0)


def test_none_to_one(zone_example_grid):
    mg, z = zone_example_grid

    # No zones exist at time 0.

    se = SpeciesEvolver(mg)
    sc = ZoneController(mg, zone_func)
    taxa = sc.populate_zones_uniformly(1)
    se.track_taxa(taxa)

    np.testing.assert_equal(len(sc.zones), 0)

    expected_df = pd.DataFrame(
        {
            "time": [0],
            "zones": [0],
            "fragmentations": [np.nan],
            "captures": [np.nan],
            "area_captured_sum": [np.nan],
            "area_captured_max": [np.nan],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    # Create a zone for time 1.

    z[[9, 10, 11, 12]] = 1

    sc.run_one_step(1)
    se.run_one_step(1)

    np.testing.assert_equal(len(sc.zones), 1)
    np.testing.assert_equal(sc.zones[0]._conn_type, zn.Connection.NONE_TO_ONE)

    expected_df = pd.DataFrame(
        {
            "time": [0, 1],
            "zones": [0, 1],
            "fragmentations": [np.nan, 0],
            "captures": [np.nan, 0],
            "area_captured_sum": [np.nan, 0],
            "area_captured_max": [np.nan, 0],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    np.testing.assert_equal(se.record_data_frame.taxa.sum(), 0)


def test_one_to_none(zone_example_grid):
    mg, z = zone_example_grid

    # Create a zone for time 0.

    z[[9, 10, 11, 12]] = 1

    se = SpeciesEvolver(mg)
    sc = ZoneController(mg, zone_func)
    taxa = sc.populate_zones_uniformly(1)
    se.track_taxa(taxa)

    np.testing.assert_equal(len(sc.zones), 1)
    zone = sc.zones[0]

    expected_df = pd.DataFrame(
        {
            "time": [0],
            "zones": [1],
            "fragmentations": [np.nan],
            "captures": [np.nan],
            "area_captured_sum": [np.nan],
            "area_captured_max": [np.nan],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    # No zones for time 1.

    z[:] = 0

    sc.run_one_step(1)
    se.run_one_step(1)

    np.testing.assert_equal(len(sc.zones), 0)
    np.testing.assert_equal(zone._conn_type, zn.Connection.ONE_TO_NONE)

    expected_df = pd.DataFrame(
        {
            "time": [0, 1],
            "zones": [1, 0],
            "fragmentations": [np.nan, 0],
            "captures": [np.nan, 0],
            "area_captured_sum": [np.nan, 0],
            "area_captured_max": [np.nan, 0],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    np.testing.assert_equal(se.record_data_frame.taxa.sum(), 1)


def test_one_to_one(zone_example_grid):
    mg, z = zone_example_grid

    # Create a zone for time 0.

    z[[9, 10, 11, 12]] = 1

    se = SpeciesEvolver(mg)
    sc = ZoneController(mg, zone_func)
    taxa = sc.populate_zones_uniformly(1)
    se.track_taxa(taxa)

    np.testing.assert_equal(len(sc.zones), 1)
    zone = sc.zones[0]

    expected_df = pd.DataFrame(
        {
            "time": [0],
            "zones": [1],
            "fragmentations": [np.nan],
            "captures": [np.nan],
            "area_captured_sum": [np.nan],
            "area_captured_max": [np.nan],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    # Modify elevation, although  there is still one zone in time 1.

    z[[11, 12]] = 0

    sc.run_one_step(1)
    se.run_one_step(1)

    np.testing.assert_equal(len(sc.zones), 1)
    np.testing.assert_equal(zone, sc.zones[0])
    np.testing.assert_equal(sc.zones[0]._conn_type, zn.Connection.ONE_TO_ONE)

    expected_df = pd.DataFrame(
        {
            "time": [0, 1],
            "zones": [1, 1],
            "fragmentations": [np.nan, 0],
            "captures": [np.nan, 0],
            "area_captured_sum": [np.nan, 0],
            "area_captured_max": [np.nan, 0],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    np.testing.assert_equal(len(se.get_extant_taxon_objects(time=1)), 1)


def test_one_to_many(zone_example_grid):
    mg, z = zone_example_grid

    # Create a zone for time 0.

    z[[9, 10, 11, 12]] = 1

    se = SpeciesEvolver(mg)
    sc = ZoneController(mg, zone_func)
    taxa = sc.populate_zones_uniformly(1)
    se.track_taxa(taxa)

    expected_df = pd.DataFrame(
        {
            "time": [0],
            "zones": [1],
            "fragmentations": [np.nan],
            "captures": [np.nan],
            "area_captured_sum": [np.nan],
            "area_captured_max": [np.nan],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    np.testing.assert_equal(len(se.get_extant_taxon_objects(time=0)), 1)

    # Break the zone in two for time 1.

    z[11] = 0

    sc.run_one_step(1)
    se.run_one_step(1)

    np.testing.assert_equal(len(sc.zones), 2)
    np.testing.assert_equal(
        {z._conn_type for z in sc.zones}, {None, zn.Connection.ONE_TO_MANY}
    )

    expected_df = pd.DataFrame(
        {
            "time": [0, 1],
            "zones": [1, 2],
            "fragmentations": [np.nan, 2],
            "captures": [np.nan, 0],
            "area_captured_sum": [np.nan, 0],
            "area_captured_max": [np.nan, 0],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    np.testing.assert_equal(len(se.get_extant_taxon_objects()), 2)


def test_many_to_one(zone_example_grid):
    mg, z = zone_example_grid

    # Create two zones for time 0.

    z[[8, 9, 10, 12]] = 1

    se = SpeciesEvolver(mg)
    sc = ZoneController(mg, zone_func)
    taxa = sc.populate_zones_uniformly(1)
    se.track_taxa(taxa)

    expected_df = pd.DataFrame(
        {
            "time": [0],
            "zones": [2],
            "fragmentations": [np.nan],
            "captures": [np.nan],
            "area_captured_sum": [np.nan],
            "area_captured_max": [np.nan],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    np.testing.assert_equal(len(se.get_extant_taxon_objects(time=0)), 2)

    # Modify elevation such that two zones each overlap the original two zones.

    z[11] = 1

    sc.run_one_step(1)
    se.run_one_step(1)

    np.testing.assert_equal(len(sc.zones), 1)
    np.testing.assert_equal(sc.zones[0]._conn_type, zn.Connection.MANY_TO_ONE)

    expected_df = pd.DataFrame(
        {
            "time": [0, 1],
            "zones": [2, 1],
            "fragmentations": [np.nan, 0],
            "captures": [np.nan, 1],
            "area_captured_sum": [np.nan, 12],
            "area_captured_max": [np.nan, 12],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    np.testing.assert_equal(len(se.get_extant_taxon_objects(time=1)), 2)


def test_many_to_many(zone_example_grid):
    mg, z = zone_example_grid

    # Create two zones for time 0.

    z[[10, 12, 17, 19, 24, 26]] = 1

    se = SpeciesEvolver(mg)
    sc = ZoneController(mg, zone_func)
    taxa = sc.populate_zones_uniformly(1)
    se.track_taxa(taxa)

    expected_df = pd.DataFrame(
        {
            "time": [0],
            "zones": [2],
            "fragmentations": [np.nan],
            "captures": [np.nan],
            "area_captured_sum": [np.nan],
            "area_captured_max": [np.nan],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    np.testing.assert_equal(len(se.get_extant_taxon_objects(time=0)), 2)

    # Modify elevation such that two zones each overlap the original two zones.

    z[[17, 19]] = 0
    z[[11, 25]] = 1

    sc.run_one_step(1)
    se.run_one_step(1)

    np.testing.assert_equal(len(sc.zones), 2)
    for zone in sc.zones:
        np.testing.assert_equal(zone._conn_type, zn.Connection.MANY_TO_MANY)

    expected_df = pd.DataFrame(
        {
            "time": [0, 1],
            "zones": [2, 2],
            "fragmentations": [np.nan, 0],
            "captures": [np.nan, 2],
            "area_captured_sum": [np.nan, 24],
            "area_captured_max": [np.nan, 12],
        }
    )
    pd.testing.assert_frame_equal(sc.record_data_frame, expected_df, check_like=True)

    np.testing.assert_equal(len(se.get_extant_taxon_objects()), 4)


def test_one_to_many_to_one(zone_example_grid):
    mg, z = zone_example_grid

    z[[9, 10, 11, 12]] = 1

    se = SpeciesEvolver(mg)
    sc = ZoneController(mg, zone_func)
    taxa = sc.populate_zones_uniformly(1, time_to_allopatric_speciation=1)
    se.track_taxa(taxa)

    z[11] = 0
    sc.run_one_step(1)
    se.run_one_step(1)

    z[11] = 1
    sc.run_one_step(1)
    se.run_one_step(1)

    np.testing.assert_equal(len(se.get_extant_taxon_objects()), 1)


def test_min_area(zone_example_grid):
    mg, z = zone_example_grid

    # Create a zone for time 0.

    z[[9, 11, 12]] = 1

    sc = ZoneController(mg, zone_func, minimum_area=5)
    np.testing.assert_equal(len(sc.zones), 1)


def test_neighborhood_structure(zone_example_grid):
    mg, z = zone_example_grid

    z[[10, 16]] = 1

    sc = ZoneController(mg, zone_func)
    np.testing.assert_equal(len(sc.zones), 1)

    sc = ZoneController(mg, zone_func, neighborhood_structure="D4")
    np.testing.assert_equal(len(sc.zones), 2)

    np.testing.assert_raises(
        ValueError, ZoneController, mg, zone_func, neighborhood_structure="D"
    )


def test_zone_func_kwargs(zone_example_grid):
    mg, z = zone_example_grid
    z[mg.core_nodes] = 1
    sc = ZoneController(mg, zone_func_with_vars, var1=1, var2=2)

    expected_mask = np.array(
        [
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
        ]
    )
    np.testing.assert_array_equal(sc.zones[0].mask, expected_mask)

    z[-2:] = 0
    expected_mask[-2:] = False

    sc.run_one_step(10)

    np.testing.assert_array_equal(sc.zones[0].mask, expected_mask)


def test_zone_taxon_range_mask(zone_example_grid):
    mg, z = zone_example_grid

    # Create a zone for time 0.

    ids = [9, 10, 11, 12]
    z[ids] = 1

    se = SpeciesEvolver(mg)
    sc = ZoneController(mg, zone_func)
    taxa = sc.populate_zones_uniformly(1)
    se.track_taxa(taxa)

    expected_mask = np.zeros(mg.number_of_nodes, bool)
    expected_mask[ids] = True
    np.testing.assert_array_equal(taxa[0].range_mask, expected_mask)

    # Remove extent so taxa range mask is all False.

    z[ids] = 0
    sc.run_one_step(1)
    se.run_one_step(1)

    expected_mask = np.zeros(mg.number_of_nodes, bool)
    np.testing.assert_array_equal(taxa[0].range_mask, expected_mask)


def test_time_to_allopatric_speciation(zone_example_grid):
    mg, z = zone_example_grid

    # Create a zone for time 0.

    z[[9, 10, 11, 12]] = 1

    se = SpeciesEvolver(mg)
    sc = ZoneController(mg, zone_func)
    taxa = sc.populate_zones_uniformly(1, time_to_allopatric_speciation=20)
    se.track_taxa(taxa)

    z[[11]] = 0

    while len(se.taxa_data_frame) == 1:
        sc.run_one_step(10)
        se.run_one_step(10)

    expected_df = pd.DataFrame(
        {
            "pid": [np.nan, 0],
            "type": 2 * [ZoneTaxon.__name__],
            "t_first": [0, 30],
            "t_final": 2 * [np.nan],
        },
        index=[0, 1],
    )
    expected_df.index.name = "tid"
    expected_df["pid"] = expected_df["pid"].astype("Int64")
    expected_df["t_final"] = expected_df["t_final"].astype("Int64")

    pd.testing.assert_frame_equal(se.taxa_data_frame, expected_df, check_like=True)


def test_pseudoextinction(zone_example_grid):
    mg, z = zone_example_grid

    z[[9, 10, 11, 12]] = 1

    se = SpeciesEvolver(mg)
    sc = ZoneController(mg, zone_func)
    taxa = sc.populate_zones_uniformly(1, persists_post_speciation=False)
    se.track_taxa(taxa)

    z[11] = 0
    sc.run_one_step(1)
    se.run_one_step(1)

    expected_df = pd.DataFrame(
        {
            "pid": [np.nan, 0, 0],
            "type": 3 * [ZoneTaxon.__name__],
            "t_first": [0, 1, 1],
            "t_final": [1, np.nan, np.nan],
        },
        index=[0, 1, 2],
    )
    expected_df.index.name = "tid"
    expected_df["pid"] = expected_df["pid"].astype("Int64")
    expected_df["t_final"] = expected_df["t_final"].astype("Int64")

    pd.testing.assert_frame_equal(se.taxa_data_frame, expected_df, check_like=True)

    expected_df = pd.DataFrame(
        {
            "time": [0, 1],
            "taxa": [1, 2],
            "speciations": [np.nan, 2],
            "extinctions": [np.nan, 0],
            "pseudoextinctions": [np.nan, 1],
        }
    )
    pd.testing.assert_frame_equal(se.record_data_frame, expected_df, check_like=True)



import os

from ..core.messages import deprecation_message


class DeprecatedSignature(DeprecationWarning):
    msg = "You are using a deprecated calling signature."

    def __init__(self, name, old=None, new=None):
        self._name = name
        self._old = old
        self._new = new

        if old:
            self._old = self._construct_call(name, self._old[0], self._old[1])
        if new:
            self._new = self._construct_call(name, self._new[0], self._new[1])

    @staticmethod
    def _construct_call(name, args, kwds):
        signature = ", ".join(
            [repr(arg) for arg in args] + [f"{k}={repr(v)}" for k, v in kwds.items()]
        )
        return f"{name}({signature})"

    def __str__(self):
        if self._new:
            use = f">>> grid = {self._new}"
        else:
            use = None

        return os.linesep + deprecation_message(self.msg, use=use)
